# =========================================================================
#
# Tracefile routines
#
# we only read and check traces as long as maininfo(tracefile) exists
# some more general routines are always enabled
#

proc uttrace_file_open {} {
	global maininfo
	if {[info exists maininfo(tracefile)]} {
		if {[catch {open $maininfo(tracefile) r} fileId]} {
			fatal "Cannot open tracefile $maininfo(tracefile): $fileId\n"
		} else {
			send_error "Opened trace file $maininfo(tracefile)\n"
			set maininfo(tracefileId) $fileId
		}
	}
}

proc uttrace_file_close {} {
	global maininfo
	if {[info exists maininfo(tracefileId)]} {
		if [catch {close $maininfo(tracefileId)} fileId] {
			fatal "Problems closing tracefile file $maininfo(tracefileId): $fileId\n"
		}
		catch {unset maininfo(tracefileId)}
		send_error "Closed trace file $maininfo(tracefile)\n"
	}
}

# next two routines are needed by the Adaptor,
# by the old <gate,experiments>
# based routines in main_util.tcl and coding.tcl
# NOTE: the Driver no longer needs these routines,
# and thus the Driver version of this file no longer
# has them (there they are commented out)
# TODO: clean this up
proc uttrace_parse_event {s} {
	set event [split $s {!}]
	set nevent {}
	foreach e $event {
		lappend nevent [string trim $e]
	}
	set gate [lindex $nevent 0]
	set exp [lrange $nevent 1 end]
	return [list $gate $exp]
}
proc uttrace_unparse_event {ev} {
	set pr_ev [lindex $ev 0]
	foreach e [lindex $ev 1] {
		append pr_ev " ! $e"
	}
	return $pr_ev
}

proc uttrace_read_event {} {
	global maininfo
	if {[info exists maininfo(tracefileId)]} {
		while {[gets $maininfo(tracefileId) line] >= 0} {
			# skip lines that start with CONCRETE
			# accept lines of form: "ABSTRACT $maininfo(nrofstepsdone) $side($chan:$pco)? ($res)"
			if {[regexp "^\[\t \]*ABSTRACT\[\t \]+(\[0-9\]+)\[\t \]+(input|output)(\\(\[^\t \\(\\)\]+\\))?\[\t \]+\\((.*)\\)(\[\t \]+\\((\[^\\(\\)\]*)\\))\[\t \]*$" $line match nr side chan_pco_opt gexps susp_opt susp]} {
				# puts stderr "A_DEBUG TRACE <$nr><$side><$chan_pco_opt><$gexps><$susp_opt><$susp>"
				set repr [utlog_tostring_event_suspension $gexps $susp]
				send_user "Trace: $side $chan_pco_opt $repr\n" ;# used by GUI
				if {[regexp "^\\((\[^\t :\\(\\)\]+)(:(\[^\t \\(\\)\]+))?\\)$" $chan_pco_opt m chan pco]} {
					set channel $chan
				} else {
					set channel [default_channel $side]
				}
	    		set ev [string trim $gexps]
				# send_user "Trace: ($side) ($channel) ($ev) ($susp)\n"
				return [list $side $channel $ev $susp]
			} elseif {[regexp "^\[\t \]*ABSTRACT\[\t \]+" $line match]} {
				puts stderr "Warning: internal error: could not match (skipping) trace line: <$line>"
			}
		}
		send_user "End of trace reached\n"
		uttrace_file_close
	}
	return [list eot]
}

proc uttrace_check_event {ev tr} {
	global maininfo
	if {[info exists maininfo(tracefileId)]} {
		if {[string compare $tr ""] != 0} {
			if {![eq_event $ev $tr]} {
				send_user "$maininfo(nrofstepsdone) Trace mismatch (expected: $tr)\n"
				uttrace_file_close
			}
		}
	}
}
proc uttrace_check_iokind {io tr} {
	global maininfo
	if {[info exists maininfo(tracefileId)]} {
		if {[llength $tr] > 0} {
			if {([string compare [lindex $io 0] [lindex $tr 0]] != 0) ||
			    ([string compare [lindex $io 1] [lindex $tr 1]] != 0)} {
				send_user "$maininfo(nrofstepsdone) Trace mismatch (expected: $tr event)\n"
				uttrace_file_close
			}
		}
	}
}

# =========================================================================
#
# More 'general' routines - do not depend on maininfo(tracefileId)
#

proc uttrace_iokind {tr_event} {
	switch -- [lindex $tr_event 0] {
		  input -
		  output {
			set result [list [lindex $tr_event 0] [lindex $tr_event 1]]
		} default {
			set result {}
		}
	}
	return $result
}

proc uttrace_channel {tr_event} {
	switch -- [lindex $tr_event 0] {
		input -
		output {
			set result [lindex $tr_event 1]
		} default {
			set result {}
		}
	}
	return $result
}

proc uttrace_event {tr_event} {
	switch -- [lindex $tr_event 0] {
		input -
		output {
			set result [lindex $tr_event 2]
		} default {
			set result {}
		}
	}
	return $result
}

proc uttrace_supension {tr_event} {
	switch -- [lindex $tr_event 0] {
		input -
		output {
			set result [lindex $tr_event 3]
		} default {
			set result {}
		}
	}
	return $result
}

# =========================================================================
