# =========================================================================
#
# start spec module and connect to it
#

proc start_spec {} {
	global timeout maininfo specinfo sidinfo
	set timeout $maininfo(timeout)
	match_max -d 10000
	
	#if {[catch {spawn -pty} foo]} {
	#	fatal "spawn for stderr of $specinfo(program) fails: $foo"
	#}
	#set sidinfo(spec,stderr) $spawn_id
	#set specinfo(pty,stderr) $spawn_out(slave,name)
	#send_error "DEBUG: SPEC Stderr match_max=[match_max]\n"
	#set sidinfo(sid,name,$sidinfo(spec,stderr)) SPEC
	#lappend sidinfo(bg) $sidinfo(spec,stderr)
	
	#generate_spec_command $specinfo(pty,stderr)
	
	generate_spec_command
	debugLvl 0 "$specinfo(cmd)\n"
	catch { send_error "DEBUG: Starting spec: ($specinfo(cmd))\n" }
	if {[catch $specinfo(cmd) foo]} {
		fatal "spawn for tester fails: $foo\n"
	}
	set sidinfo(spec) $spawn_id
	set specinfo(pid) [exp_pid -i $sidinfo(spec)]
	send_error "DEBUG: SPEC match_max=[match_max]\n"
	set sidinfo(sid,name,$sidinfo(spec)) "SPEC"
	io_add_peer spec $sidinfo(spec) $specinfo(timeout)
	io_config $maininfo(timeout)
	
	# What do we do when we get a A_GUIDANCE_VERDICT ???
	# In principle, the user should decide, for example, by giving a
	#       on guidance-verdict verdict command
	# command.
	# Currently this is not yet possible.
	# Currently, we simply stop when we get a A_GUIDANCE_VERDICT command.
	#
	# Actually, we could try to rewrite the command-interpreter such
	# that each command is mapped on a tcl command, and the
	# command-interpreter simple interprets these tcl commands.
	# (Effectively, the command-interpreter simple feeds the commands that
	#  it gets directly(?) to the (safe-?) tcl interpreter. )
	# It would then be simple to allow 'scripting', like
	#	source myscript.tcl
	#	on guidance-verdict verdict mycommand-from-myscript
	# However, we then have to make sure that the arguments that we
	# want to give to commands are 'valid' tcl arguments, and don't
	# get treated in a special way.
	
#	expect_background {
#		-i $sidinfo(spec,stderr) full_buffer {
#			send_error "A_WARNING SPEC Stderr Buffer Overflow!!!\n"
#			set txt $expect_out(buffer)
#			send_error "A_WARNING SPEC Stderr Buffer Contents Skipped: $txt\n"
#			send_error "A_ERROR SPEC Stderr Buffer Contents Skipped: $txt\n"
#			exit
#		}
#		-i $sidinfo(spec,stderr) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
#			set txt $expect_out(1,string)
#			send_error "SPEC Debug: $txt\n"
#		}
#		-i $sidinfo(spec,stderr) -re "(\[^\r\n]*)\r*\n" {
#			set l $expect_out(1,string)
#			send_error "SPEC Stderr: $l\n"
#		}
#	}
	
	expect_before {
		-i $sidinfo(spec) full_buffer {
			warning "SPEC: Buffer Overflow!!!\n"
			set txt $expect_out(buffer)
			warning "SPEC: Buffer Contents Skipped: $txt\n"
			fatal "SPEC: Buffer Contents Skipped: $txt\n"
			exit
		}
		-i $sidinfo(spec) eof {
			global sidinfo
			if {[info exists sidinfo(spec)]} {
				unset sidinfo(spec)
			}
			io_rm_peer spec
			if {![cleaning_up]} {
				fatal "End-of-file from specification\n"
			}
		}
	}
	expect_after {
		-i $sidinfo(spec) -re "^A_GUIDANCE_VERDICT(\[ \t]\[\t ]*\[^\r\n]\[^\r\n]*)?\[ \t]*\r*\n" {
			global specinfo maininfo
			set txt $expect_out(1,string)
			switch -- [lindex $txt 0] \
				INCONSISTENT {
					warning "SPEC: inconsistency detected in pattern\n"
					set specinfo(guidance,inconsistent) 1
					set maininfo(stopnow) 1
					set maininfo(verdict) inconsistent
				} INCONCLUSIVE {
					set patterns [lrange $txt 1 end]
					
					warning "SPEC: pattern inconclusive $patterns\n"
					set specinfo(guidance,inconclusive) 1
					set maininfo(stopnow) 1
					set maininfo(verdict) inconclusive
				} HIT {
					warning "SPEC: pattern end-of-test detected\n"
					set specinfo(guidance,endoftest) 1
					set maininfo(stopnow) 1
					set maininfo(verdict,guidance) hit
					if {[string compare $maininfo(verdict) fail] != 0} {
						set maininfo(verdict) pass
					}
				} MISS {
					warning "SPEC: pattern end-of-test detected\n"
					set specinfo(guidance,endoftest) 1
					set maininfo(stopnow) 1
					set maininfo(verdict,guidance) miss
					if {[string compare $maininfo(verdict) fail] != 0} {
						set maininfo(verdict) pass
					}
				} PASS {
					warning "SPEC: pattern end-of-test detected\n"
					set specinfo(guidance,endoftest) 1
					set maininfo(stopnow) 1
					set maininfo(verdict) pass
					# set maininfo(verdict,guidance) hit
				}
			exp_continue
		}
		-i $sidinfo(spec) -re "^C_\[^\n]*\n" {
			# skip. This is echo'ed back from the specification module
			exp_continue
		}
		-i $sidinfo(spec) -re "^A_QUIT\r*\n" {
			set result [list error]
		}
		-i $sidinfo(spec) -re "^A_WARNING(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			warning "SPEC: $txt\n"
			exp_continue
		}
		-i $sidinfo(spec) -re "^A_MSG(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			message "SPEC: $txt\n"
			exp_continue
		}
		-i $sidinfo(spec) -re "^A_NOTIMPLEMENTED(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			nonfatal "SPEC unimplemented command: $txt\n"
			set result [list error ]
		}
		-i $sidinfo(spec) -re "^A_ERROR\[ \t]\[\t ]*UnknownCommand:\[ \t]\[\t ]*(\[^\r\n]\[^\r\n]*)\[ \t]*\r*\n" {
			# This regexp tries to catch the 'not-implemented'
			# answers from old spec components (i.e. those that do not
			# yet implement the A_NOTIMPLEMENTED answer)
			# to allow us to continue working, instead of interpreting
			# it as a fatal error.
			# WARNING: this regexp 'knows' that old spec components
			# use this particular format in case of unknown commands
			# (this works at least for the LOTOS and Promela components):
			#   A_ERROR UnknownCommand: command-that-was-not-implemented
			# (including the :)
			# This just might fail for other specification componts that
			# use a slightly different error message...
			set txt $expect_out(1,string)
			warning "command not implemented in SPEC: $txt\n"
			set result [list error ]
		}
		-i $sidinfo(spec) -re "^A_ERROR(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			fatal "Error SPEC: $txt\n"
		}
		-i $sidinfo(spec) -re "^A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			send_error "SPEC Debug: $txt\n"
			exp_continue
		}
		-i $sidinfo(spec) -re "^A_LOG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			send_error "SPEC Log: $txt\n"
			utlog_tofile_string SPEC $txt
			exp_continue
		}
		-i $sidinfo(spec) -re "^(A_\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			warning "Unexpected SPEC Answer Ignored : $txt\n"
			exp_continue
		}
		-i $sidinfo(spec) -re "^STDERR:A_DEBUG\[ \t]\[ \t]*(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			send_error "SPEC Debug: $txt\n"
			exp_continue
		}
		-i $sidinfo(spec) -re "^STDERR:(\[^\r\n]*)\r*\n" {
			set txt $expect_out(1,string)
			send_error "SPEC Stderr: $txt\n"
			exp_continue
		}
		-i $sidinfo(spec) -re "^(\[^\r\n]*)\r*\n" {
			set l $expect_out(buffer)
			if {[regexp -- "^(\[^\r\n\]*)" $l match sub]} {
				send_error "SPEC: $sub\n"
			} else {
				send_error "SPEC: $l\n"
			}
			exp_continue
		}
	}
}

# =========================================================================
