#! @PERL@ --

#
# usage: unhexify
# 
# DECRIPTION:
#
# reads hex on stdin, writes binary on stdout
#

use strict;

require 5.002;
use Getopt::Long;

my ( $prog, $BUFSIZ,
     $rin, $rout, $win, $wout, $ein, $eout,
     %optcl, $result, $debug,
     $command, $buffer, %buf, $nfound, $nread,
     $ascii
   );

$BUFSIZ = 1024;

%optcl = ();
$result = GetOptions(
			\%optcl,
			"debug:i"
		);
die "usage: $0\n"
	unless $result;

$debug = 1 if ((defined $optcl{"debug"}) && ! $optcl{"debug"});
$debug = $optcl{"debug"} if ((defined $optcl{"debug"}) && $optcl{"debug"});

if ($debug) {
	print $prog.": ($result)(".$optcl{"debug"}.")\n";
	print $prog.": ($result)(".$debug.")\n";
}
exit if $debug >= 3;

$| = 1;				# don't buffer output on stdout

$rin = $win = $ein = "";
$rin = fdbits(fileno(STDIN));
$ein = $rin | $win;

$buf{'STDIN'} = "";

for (;;) {
	print STDERR "fileDescs:  ", &printVec($rin), "\n"
	    if $debug;
	if (($nfound = select($rout=$rin, $wout=$win, $eout=$ein, undef)) < 0) {
	    die "$prog: select: $!";
	}

	print STDERR "rout:  ", &printVec($rout), ", " if $debug;
	
	if (vec($rout, fileno(main::STDIN), 1)) {
		print STDERR "STDIN ready\n" if $debug;
		$buffer = '';
		if (($nread = sysread(main::STDIN, $buffer, $BUFSIZ)) > 0) {
		    $buf{'STDIN'} .= $buffer;
		    print STDERR "STDIN read $nread\n" if $debug;
		} elsif ($nread == 0) {
		    print STDERR "STDIN eof\n" if $debug;
		    exit(0);
		}
		if ($buf{'STDIN'} =~ /^([^\n]*)\n/) {
			$command = $1;
			$buf{'STDIN'} = $';
			$command =~ s/[\ \t]//g;
			$ascii = pack("H*", $command);
			$ascii .= "\n"
				unless ($ascii =~ /\n$/);
			print STDOUT $ascii;
		}
	}
}


# some support routines for the select code.
sub printVec {
    my($v) = @_;
    my($i, $result);

    for ($i = (8*length($v)) - 1; $i >= 0; $i--) {
	$result .= (vec($v, $i, 1)) ? "1" : "0";
    }
    $result;
}

sub fhbits {
	my(@fhlist) = @_;
	my($bits);
	for (@fhlist) {
		vec($bits, fileno($_), 1) = 1;
	}
	return $bits;
}
sub fdbits {
	my(@fdlist) = @_;
	my($bits);
	for (@fdlist) {
		vec($bits, $_, 1) = 1;
	}
	return $bits;
}
