#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "parser.h"
#include "set.h"
#include "krfsm.h"
#include "expui.h"
#include "pfunctions.h"
#include "main.h"
#include "./version.h"

/* global */
FILE *fo,*fi,*fe,*o;
FILE *f;		/* input file */
char *prog;		/* name of the executable */
char filename[100];	/* name of input file */
int nr_errs;

/* executable options */
int cmd_verbose=0;	/* print as much logging as possible */
int cmd_debug=0;	/* print debugging output */
int cmd_list=0;		/* execute the list command */
int cmd_ptree=0;        /* generate parse tree */
int cmd_rfmsgraph=0;    /* generate rfsm graph */
int cmd_explore=1;    /* explore generated rfsm graph */
int cmd_todot=0;    /* generate rfsm dot graph on stdout */
int cmd_output_strings=1;    /* output strings instead of id's */


/***********************/

boolean
cfg_output_strings(void)
{
	return (cmd_output_strings ? True : False);
}

void
dfprintf(FILE *out, char *fmt, ... )
{
	va_list args;

	va_start(args, fmt);
	if (cmd_debug)
		(void)vfprintf(out, fmt, args);
	va_end(args);
	fflush(out);
}

void
print_verbose(FILE *out, char *fmt, ... )
{
        va_list args;

        va_start(args, fmt);
	if (cmd_verbose) (void)vfprintf(out, fmt, args);
        va_end(args);
        fflush(out);
}

int
main (int argc, char *argv[])
{
	char buf[100];
	char *epsilon_label = "EPSILON";
	fo=stdout;
	fi=stdin;
	fe=stderr;
  
	prog = argv[0];
	argc--;
	argv++;

	/* parse options */
	while (argc > 0 && argv[0][0] == '-') {
		switch (argv[0][1]) {
		case 'a':
			cmd_output_strings = 0;
			break;
		case 'd':
			cmd_todot = 1;
			cmd_explore = 0;
			break;
		case 'e':
			cmd_explore = 1;
			cmd_todot = 0;
			break;
		case 'l': if (argc > 1) {
				epsilon_label = argv[1];
				argc--;
				argv++;
			} else {
				fprintf(fe,"%s: argument missing for %s\n",prog,argv[0]);
				exit(1);
			}
			break;
		case 'L':
			epsilon_label = 0;
			break;
		case 'p':
			cmd_ptree = 1;
			break;
		case 'r':
			cmd_rfmsgraph = 1;
			break;
		case 's':
			cmd_output_strings = 1;
			break;
		case 'v':
			cmd_verbose =  1;
			break;
		case 'h':
			fprintf(fe,"%s %s\n",T_Name, T_Version);
			fprintf(fe,"usage: %s [-options] [filename]\n",prog);
			fprintf(fe,"available options:\n");
			fprintf(fe,"\t-a     : output action id's instead of strings\n");
			fprintf(fe,"\t-d     : print RFSM on stdout\n");
			fprintf(fe,"\t-e     : explore RFSM (on stdin/stdout)(default)\n");
			fprintf(fe,"\t-h     : print this help\n");
			fprintf(fe,"\t-l eps : add epsilon loop on label eps to accept state (default: add loop on label EPSILON)\n");
			fprintf(fe,"\t-L     : do not add epsilon loop to accept state\n");
			fprintf(fe,"\t-p     : print parsetree in <dest>.pt.dot\n");
			fprintf(fe,"\t-r     : print RFSM in <dest>.rfsm.dot\n");
			fprintf(fe,"\t-s     : output strings instead of action id's (default)\n");
			fprintf(fe,"\t-v     : enforce verbose mode\n");
			exit(1);
			break;
		default:
			fprintf(fe,"%s: unknown option %s\n",prog,argv[0]);
			exit(1);
			break;
		}
		argc--;
		argv++;
	}
    if (argc >1) {
    	fprintf(fe,"use: %s [-option] ... [-option] [filename]\n",prog);
		exit(1);
    } else if (argc == 0)  /* using it as filter ? */ {
		f = stdin;
		strcpy(filename,"stdin");
    } else {
    	strcpy(filename,argv[0]);
		if (!(f = fopen(argv[0], "r"))) {
			fprintf (fe, "%s: Unable to open file \"%s\"\n",
				prog, argv[0]);
            exit (1);
		}
	}
	lexerinit();
	yyin=f;    
	nr_errs = 0;
	/***** start */
	print_verbose(fe,"%s %s\n",T_Name,T_Version);
	print_verbose(fe,"Parsing specification\n");
	yyparse();
	fclose(f);
	if (nr_errs) {
	  fprintf (fe,"Terminating: errors found during parsing\n");
	  exit(1);
	}
	
	/* extend parse tree with hit action, if we have one */
	if (epsilon_label) {
		decl eps = Action(Id(Uid(mkcasestring("-- epsilon --"))),
		                  mkcasestring(epsilon_label));
		epsilon_set(eps);
		parsetree = add_epsilon(parsetree, eps);
	}
	
	if (cmd_ptree) { /* print parsetree */	  
		print_verbose(fe,"Printing parsetree to file: %s.pt.dot",
			filename);
		strcpy(buf,filename);
		strcat(buf,".pt.dot");
		if (!(f = fopen(buf, "w"))) {
			fprintf (fe, "%s: Unable to open file \"%s.pt.dot\"\n",
				prog, filename);
			exit(1);
		}
		fprintdot_spec(f, parsetree,buf,0,0,True,True,True);
		fclose(f);
		print_verbose(fe,"Printing parsetree to file: %s.pt.txt",
			filename);
		strcpy(buf,filename);
		strcat(buf,".pt.txt");
		if (!(f = fopen(buf, "w"))) {
			fprintf (fe, "%s: Unable to open file \"%s.pt.txt\"\n",
				prog, filename);
			exit(1);
		}
		fprint_spec(f,parsetree);
		fclose(f);

	}

	unparse_spec(parsetree,dummyprint,p_decl); /* create rfsm of spec */

	if (cmd_rfmsgraph) { /* print graph of rfms */	  
		print_verbose(fe,"Printing parsetree to file: %s.rfsm.dot",
			filename);
		strcpy(buf,filename);
		strcat(buf,".rfsm.dot");
		if (!(f = fopen(buf, "w"))) {
			fprintf (fe, "%s: Unable to open file \"%s\"\n",
				prog, buf);
			exit(1);
		}
		PrintObjRFSM(f, parsetree);
		fclose(f);
	}
	if (cmd_todot) { /* print graph of rfms */	  
		print_verbose(fe,"Printing parsetree to file: %s.rfsm.dot",
			filename);
		PrintObjRFSM(stdout, parsetree);
	}
	if (cmd_explore) {
		ui();
	}
	return 0;
}
