import java.io.*;
import java.util.*;
import lts.*;

/**
* Batch mode run of the torx interface
*/
public class TorxBatch implements LTSInput, LTSOutput, TorxOutput {

    // used to implement muCSPInput
	int fPos = -1;
	String fSrc = "\n";

	public static void main(String args[]) {
		String choice;
		String file;
		boolean todot;

		int i = 0;
		String s;
		int errors = 0;

		choice = "DEFAULT";
		todot = false;
		file = "";

		while (i < args.length) {
			s = args[i];
			i++;
			if (s.compareTo("-c") == 0) {
				if (i < args.length) {
					choice = args[i];
					i++;
				} else {
					System.err.println("A_WARNING flag -c needs argument");
					errors++;
				}
			}
			else if (s.compareTo("-d") == 0)
				todot = true;
			else if (s.compareTo("-l") == 0)
				choice = "";
			else if (s.compareTo("--") == 0) {
				break;
			}
			else if (s.startsWith("-")) {
				System.err.println("A_WARNING unknown flag: " + s);
				errors++;
			}
			else {
				i--;
				break;
			}
		}
		
		if (i < args.length) {
			file = args[i];
			i++;
		} else {
			System.err.println("A_WARNING no lts input file given");
			errors++;
		}
		while (i < args.length) {
			s = args[i];
			i++;
			System.err.println("A_WARNING spurious command line argument: " + s);
			errors++;
		}
		if (errors > 0) {
			System.err.println("A_ERROR usage: expltsa [ -d ] [-l | -c composition] file.lts");
			return;
		}
		new TorxBatch(file, choice, todot);
	}
	
	public TorxBatch (String f, String s, boolean todot) {
	  SymbolTable.init();
		try {
			FileInputStream fin = new FileInputStream(new File(f));
			BufferedReader di = new BufferedReader(new InputStreamReader(fin));
			fSrc = "";
			String line;
			while((line = di.readLine()) != null) {
				fSrc += line + "\n";
	  		}
		}	
		catch (FileNotFoundException e) {
			System.err.println("A_ERROR LTSA: cannot open " + e.getMessage());
			return;
		}
		catch (IOException e) {
			System.err.println("A_ERROR LTSA: IO exception: " + e.getMessage());
			return;
		}
		//System.err.println("Input:------------------");
		//System.err.println(fSrc);
		//System.err.println("------------------");
		Hashtable cs = parse();
		if (cs == null) return;
		String choice = "DEFAULT";
        if (cs.size()==0) {
            System.err.println("A_MSG LTSA: selectable composition: "+ choice);
        } else  {
            Enumeration e = cs.keys();
            String es;
            while(e.hasMoreElements()) {
            	es = (String)e.nextElement();
                System.err.println("A_MSG LTSA: selectable composition: "+ es);
                if (s.compareTo(es) == 0) {
                	choice = s;
                }
            }
        }
		if (s.compareTo("") == 0) {
			return;
		}
		if (s.compareTo(choice) != 0) {
			System.err.println("A_ERROR: LTSA: composition not present (use -l to list): " + s);
			return;
		}
		CompositeState current = docompile(choice);
		if (current == null) return;
		Analyser a = new Analyser(current,this,null);
		TorxLTS lts = new TorxLTS(a, f);
		if (todot) {
			lts.todot();
		} else {
			new TorxExplorer(lts, System.in, this).handleInput();
		}
	}

//------------------------------------------------------------------------
   private CompositeState  docompile(String s) {
        fPos = -1;
        CompositeState cs=null;
        LTSCompiler comp = new LTSCompiler(this,this,".");
        try {
            cs = comp.compile(s);
        } catch (LTSException x) {
            displayError(x);
        }
        return cs;
    }

    private Hashtable parse() {
        fPos = -1;
        LTSCompiler comp = new LTSCompiler(this,this,".");
        Hashtable cs = new Hashtable();
        Hashtable ps = new Hashtable();
        try {
            comp.parse(cs,ps);
        } catch (LTSException x) {
            displayError(x);
            return null;
        }
        return cs;
    }

   private void displayError(LTSException x) {
        if (x.marker!=null) {
            int i = ((Integer) (x.marker)).intValue();
            int line = 1;
            for(int j= 0; j<i; ++j) {
                if (fSrc.charAt (j) == '\n') ++line;
            }
            System.err.println("ERROR line:"+line+" - "+x.getMessage());
        } else
          System.err.println("ERROR - "+x.getMessage());
   }
   
//------------------------------------------------------------------------

	// Implementation of the DarwinEnvironment Interface

	public char nextChar () {
		fPos = fPos + 1;
		if (fPos < fSrc.length ()) {
			return fSrc.charAt (fPos);
		} else {
			//fPos = fPos - 1;
			return '\u0000';
		}
	}


	public char backChar () {
		char ch;
		fPos = fPos - 1;
		if (fPos < 0) {
			fPos = 0;
			return '\u0000';
		}
		else
			return fSrc.charAt (fPos);
	}


	public int getMarker () {
		return fPos;
	}

	public void clearOutput() {
	}
	public void outln(String s) {
		System.err.println("A_DEBUG LTSA: " + s);
	}
	public void out(String s) {
		System.err.println("A_DEBUG LTSA: " + s);
	}

//------------------------------------------------------------------------
	// Implementation of the TorxOutput Interface
	
	public void msgln ( String str ) {
		System.err.println(str);
	}
	public void answerln ( String str ) {
		System.out.println(str);
	}
	public void echoln ( String str ) {
	}
	
//------------------------------------------------------------------------

}
