import java.io.*;
import java.util.*;

public class TorxExplorer {
	TorxLTS lts;
	InputStream in;
	TorxOutput out;

	TorxExplorer(TorxLTS lts, InputStream in, TorxOutput out) {
		this.lts=lts;
		this.in=in;
		this.out=out;
	}
	
	public void handleInput() {
		try {
			BufferedReader di = new BufferedReader(new InputStreamReader(in));
			String line;
			while((line = di.readLine()) != null) {
	  			out.echoln(line);
	  			if (!handleLine(line))
	  				return;
	  		}
			out.msgln("\t-- Eof read");
		}
		catch (EOFException eof) {
			out.msgln("\t-- Eof exception");
		}
		//catch (SocketException e) {
		//	out.msgln("\t-- Socket closed");
		//}
		catch (IOException e) {
			out.msgln("\t-- Error during read from file: "
				 + e.toString());
		}
	}
	
//------------------------------------------------------------------------

	// return true if more to handle
	// return false if last line
	private boolean handleLine(String line) {
		line = line.trim();
		if (line.length() <= 0)
			return true;
		int kw = line.charAt(0);
		String s = line.substring(0,1);
		switch(kw) {
		case 'q':
	  		cmd_quit();
			return false;
		case 'r':
			cmd_reset();
			return true;
		case 'e':
			cmd_explore(line.substring(1));
			return true;
		default:
			cmd_unknown(s);
			return true;
		}
	}

	private void cmd_quit() {
	  	out.answerln("Q");
	}
		
	private void cmd_reset() {
		TorxTransition t = lts.initial();
	  	out.answerln("R" + "\t" + t.dst.nr() + "_" + t.nr() + "." + t.number() + "\t" + 1);
	}
	
	private void cmd_explore(String line) {
		StringTokenizer st = new StringTokenizer(line);
		if (! st.hasMoreTokens()) {
			out.answerln("E0 Missing event number");
			return;
		}
		String tok = st.nextToken();
		int j = tok.lastIndexOf(".");
		if (j >= 0)
			tok = tok.substring(j+1, tok.length());
		try {
			int nr = Integer.parseInt(tok);
			TorxTransition t = lts.lookupTrans(nr);
			if (t == null) {
				out.answerln("E0 Unknown event number: " + nr);
				return;
			}
			TorxState s = t.dst();
	   		out.answerln("EB");
			Iterator i = s.succ().iterator();
			while (i.hasNext()) {
				TorxTransition tt = (TorxTransition)i.next();
				out.answerln(tt.unparse());
			}
	  		out.answerln("EE");
		} catch (NumberFormatException e) {
			out.answerln("E0 ParseError: not a number: " + tok);
		}
	}

	public void cmd_unknown(String s) {
		out.answerln("A_ERROR UnknownCommand: " + s);
	}

}

//------------------------------------------------------------------------
