#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "spexDomain.h"

/*
 * Private global variables
 */
RANGE_VALUE_TYPE rangeCmpArg;	// Used by the private RHas__ functions


static DSetP *DomainStore = NULL;
static Domain lastStoredD;

/*
 * Private function prototypes
 */
char	RHasGt(DRangeP range);
char	RHasGtEq(DRangeP range);
char	RHasSt(DRangeP range);
char	RHasStEq(DRangeP range);
char	RHas(DRangeP range);
char	RHasNot(DRangeP range);

DRangeP	RGt(DRangeP range);
DRangeP	RGtEq(DRangeP range);
DRangeP	RSt(DRangeP range);
DRangeP	RStEq(DRangeP range);

// Clones the range, except when it contains the value in 'rangeCmpArg' (then it returns NULL)
DRangeP RCloneExcept(DRangeP range);

void	DCreateStore();

/* 
 * Public Domain functions
 */

// Creates a new domain from the string specified. This string has to be in the
// form of "(el), (el)", where (el) can be either "(val)" or "(val)..(val)", where
// (val) is an integer. 
Domain newDomain(const char * input) {
	
	return DStoreSet(SetMakeFromString(input));
}
 
Domain newSingleDomain(RANGE_VALUE_TYPE lb, RANGE_VALUE_TYPE ub) {
	return DStoreSet(SetMakeSingle(RangeMakeSingle(lb, ub)));
}

Domain  DomainAddOne(Domain a, RANGE_VALUE_TYPE lb, RANGE_VALUE_TYPE ub) {
	DSetP tmpSet;
	Domain retDomain;
	
	tmpSet = SetMakeSingle(RangeMakeSingle(lb, ub));
	
	retDomain = DStoreSet(SetUnion(DGetSet(a), tmpSet));
	
	SetFree(tmpSet);
	
	return retDomain;
}


// Internal functions for storing and loading Sets
Domain DStoreSet(DSetP setA) {
	Domain i;
	if (DomainStore == NULL) {
		DCreateStore();
	}
	if (! SetIsSane(setA)) {
		setA = SetSanitize(setA);
	}
	for(i = 0; i <= lastStoredD; i++) {
		if (SetEq(setA, DGetSet(i))) {
			// TODO Should free() the passed set if already stored.
			return i;
		}
	}
	
	if (lastStoredD >= STORAGE_SIZE - 1) {
		fprintf(stderr, "Error. No space available to store new Domain. Please recompile spexDomain with larger STORAGE_SIZE.\n");
		fprintf(stderr, "Please note that DOMAIN may have to be set to a larger type as well.\n");
	}
	lastStoredD +=1;
	DomainStore[lastStoredD] = setA;
	
	return lastStoredD;
}
	


// Returns the Set stored in the domain, or emptySet if no set was stored there.
DSetP  DGetSet(Domain d) {
	if (DomainStore == NULL) {
		DCreateStore();
	}
	if (d > lastStoredD) {
		fprintf(stderr, "Warning. Trying to retrieve an unknown Domain. An empty Domain is returned instead.\n");
		return DomainStore[0];
	}
	return DomainStore[d];
}

char * 	DomainToString(Domain d) {

	return SetToString(DGetSet(d));

}

// Check if domain 'a' is equal to 'b'.
char	DomainEqual(Domain a, Domain b) {
	return (a == b);
	
	// Altertively
	// return SetEq(DGetSet(a), DGetSet(b));
}

// Move the domain up- or downwards.
Domain	DomainMove(Domain d, RANGE_VALUE_TYPE value) {
	rangeShiftValue = value;
	return DStoreSet(SetMap(DGetSet(d), RangeShift));
}




char	DHasGt(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	return SetFind(DGetSet(d), RHasGt)?1:0;
}

char	DHasGtEq(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	return SetFind(DGetSet(d), RHasGtEq)?1:0;
}
char	DHasSt(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	return SetFind(DGetSet(d), RHasSt)?1:0;
}
char	DHasStEq(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	return SetFind(DGetSet(d), RHasStEq)?1:0;
}
char	DHas(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	return SetFind(DGetSet(d), RHas)?1:0;
}


// Removes all values from the domain smaller or equal to 'value'. 
// (Domain > value)
Domain	DomainGt(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	
	return DStoreSet(SetMap(DGetSet(d), RGt));
}

Domain	DomainGt2(Domain *d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	
	*d = DStoreSet(SetMap(DGetSet(*d), RGt));
	return *d;
}


// Removes all values from the domain smaller than 'value'. 
// (Domain >= value)
Domain	DomainGtEq(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	
	return DStoreSet(SetMap(DGetSet(d), RGtEq));
}

// Removes all values from the domain greater or equal to 'value'. 
// (Domain < value)
Domain	DomainSt(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	
	return DStoreSet(SetMap(DGetSet(d), RSt));
}

Domain	DomainSt2(Domain *d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	
	*d = DStoreSet(SetMap(DGetSet(*d), RSt));
	return *d;
}

// Removes all values from the domain greater than 'value'. 
// (Domain <= value)
Domain	DomainStEq(Domain d, RANGE_VALUE_TYPE value) {
	rangeCmpArg = value;
	
	return DStoreSet(SetMap(DGetSet(d), RStEq));
}

// Removes all values from the domain not equal to 'value'.
// Or rather, make a new domain with just one value.
Domain	DomainEq(Domain d, RANGE_VALUE_TYPE value) {
	return DStoreSet(SetMakeSingle(RangeMakeSingle(value, value)));
}

// Removes all values from the domain equal to 'value'.
Domain	DomainNEq(Domain d, RANGE_VALUE_TYPE value) {
	DSetP set;
	DRangeP range, range_low, range_up;
	
	set = DGetSet(d);
	rangeCmpArg = value;
	// Find the range in the set that has the element.
	range = RangeClone(SetFind(set, RHas));
	// If there is no 'value' in the domain, we're done.
	if (range == NULL) {
		return d;
	}
	
	// Remove the element that contains 'value'
	set = SetMap(set, RCloneExcept);
	
	// TODO Probably could/should be done by a RangeSplit function
	if ( range->lb < value && range->ub > value) {
		range_up = RangeClone(range);
		range_low = range;
		
		range_up->lb = value + 1;
		range_low->ub = value - 1;
		//fprintf(stderr, "Inserting '%d..%d' and '%d..%d'\n", range_low->lb, range_low->ub, range_up->lb, range_up->ub);
		set = SetInsertElem(set, range_low);
		set = SetInsertElem(set, range_up);
		return DStoreSet(set);
	} else if (range->lb == value && range->ub > value) {
		range->lb = value + 1;
		set = SetInsertElem(set, range);
	} else if (range->lb < value && range->ub == value) {
		range->ub = value - 1;
		set = SetInsertElem(set, range);
	} else if (range->lb == value && range->ub == value) {
		// Do nothing
	} else {
		// Apperantly value was not in the set... This should not
		// have happened.
		fprintf(stderr, "Warning. Could not find value '%d' in range '%d..%d'.\n", value, range->lb, range->ub);
		//exit(-1);
	}
			
	return DStoreSet(set);
}


// Removes all values not between 'lb' and 'ub' (resp. lower- and upper bound)
Domain	DomainBetween(RANGE_VALUE_TYPE lb, Domain d, RANGE_VALUE_TYPE ub) {
	// TODO Create the DomainBetween function.
	fprintf(stderr, "Warning. DomainBetween function is not implemented yet.\n");
	return 0;
}


// Removes all values from 'a' which are not in 'b'.
Domain	DomainIntersect(Domain a, Domain b) {
	return DStoreSet(SetIntersect(DGetSet(a), DGetSet(b)));
}

// Combines all values from 'a' and 'b'.
Domain	DomainUnion(Domain a, Domain b) {
	return DStoreSet(SetUnion(DGetSet(a), DGetSet(b)));
}

// Removes all values from 'a' which are in 'b'.
Domain	DomainRemove(Domain a, Domain b) {
	return DStoreSet(SetRemove(DGetSet(a), DGetSet(b)));
}


/* 
 * Private Domain functions
 */
 
void	DCreateStore() {
	DomainStore = calloc(STORAGE_SIZE, sizeof(DSetP));
	if (DomainStore == NULL) {
		printf("Error. Could not allocate space for Domain storage. Program will terminate now.\n");
		exit(EXIT_FAILURE);
	}
	lastStoredD = 0;
	DomainStore[lastStoredD] = emptySet;
}



char	RHasGt(DRangeP range) {
	return RangeGt(range, rangeCmpArg);
}
char	RHasGtEq(DRangeP range) {
	return RangeGtEq(range, rangeCmpArg);
}
char	RHasSt(DRangeP range) {
	return RangeSt(range, rangeCmpArg);
}
char	RHasStEq(DRangeP range) {
	return RangeStEq(range, rangeCmpArg);
}
char	RHas(DRangeP range) {
	return RangeContains(range, rangeCmpArg);
}
char	RHasNot(DRangeP range) {
	return ! RangeContains(range, rangeCmpArg);
}


DRangeP	RGt(DRangeP range) {
	DRangeP newRange = NULL;
	if(RangeGt(range, rangeCmpArg)) {
		newRange = RangeClone(range);
		// TODO Try to remove the + 1 here
		newRange->lb = (range->lb > rangeCmpArg + 1) ? range->lb : rangeCmpArg + 1;
	}
	return newRange;
}
DRangeP	RGtEq(DRangeP range) {
	DRangeP newRange = NULL;
	if(RangeGtEq(range, rangeCmpArg)) {
		newRange = RangeClone(range);
		newRange->lb = (range->lb > rangeCmpArg) ? range->lb : rangeCmpArg;
	}
	return newRange;
}
DRangeP	RSt(DRangeP range) {
	DRangeP newRange = NULL;
	if(RangeSt(range, rangeCmpArg)) {
		newRange = RangeClone(range);
		// TODO Try to remove the - 1 here
		newRange->ub = (range->ub < rangeCmpArg - 1) ? range->ub : rangeCmpArg - 1;
	}
	return newRange;
}
DRangeP	RStEq(DRangeP range) {
	DRangeP newRange = NULL;
	if(RangeStEq(range, rangeCmpArg)) {
		newRange = RangeClone(range);
		newRange->ub = (range->ub < rangeCmpArg) ? range->ub : rangeCmpArg;
	}
	return newRange;
}
	
// Clones the range, except when it contains the value in 'rangeCmpArg' (then it returns NULL)
DRangeP RCloneExcept(DRangeP range) {
	if (! RangeContains(range, rangeCmpArg)) {
		return RangeClone(range);
	} else {
		return NULL;
	}
}



