#ifndef SPEX_RANGE
#define SPEX_RANGE

#include <limits.h>

#ifndef RANGE_VALUE_TYPE
#define RANGE_VALUE_TYPE signed int
#define RANGE_VALUE_MAX	 INT_MAX
#define RANGE_VALUE_MIN  INT_MIN 
#endif

typedef struct d_range {
	RANGE_VALUE_TYPE lb;	// Lower bound
	RANGE_VALUE_TYPE ub;	// Upper bound
} DRange;

typedef DRange * DRangeP;

// Global variable for RangeShift function
extern RANGE_VALUE_TYPE rangeShiftValue;


DRangeP	RangeMakeSingle(RANGE_VALUE_TYPE value_lb, RANGE_VALUE_TYPE value_ub);
DRangeP *	RangeMakeArray(RANGE_VALUE_TYPE values[], int nrValues);


// Checks if r2 is contained (completely) in r1
char	RangeContainedIn(DRangeP r1, DRangeP r2);
char	RangeContains(DRangeP r1, RANGE_VALUE_TYPE value);

char	RangeEq(DRangeP, DRangeP);

// Compare ranges. Will return 0 if the ranges are equal, negative if r1 is smaller than r2, or
// positive otherwise.
// If the lb of r1 and r2 is equal, the upper bounds are compared. Otherwise, the lowerbounds
// are compared.
int	RangeCmp(DRangeP r1, DRangeP r2);
// Will return true if r1 is smaller than r2 => (r1->ub < r2->lb)
int	RangeSmallerThan(DRangeP r1, DRangeP r2);
// Will return true if r1 is larger than r2 => (r1->lb > r2->ub)
int	RangeLargerThan(DRangeP r1, DRangeP r2);

char	RangeHasOverlap(DRangeP, DRangeP);
DRangeP	RangeOverlap(DRangeP, DRangeP);

char	RangeCanCombine(DRangeP, DRangeP);
DRangeP RangeCombine(DRangeP, DRangeP);

DRangeP * RangeDelete(DRangeP r1, DRangeP r2);

// Moves the range up by 'amount'. If 'amount' is negative, the range will be shifted
// downwards.
DRangeP RangeMove(DRangeP range, RANGE_VALUE_TYPE amount);

// Returns a new range, shifted up by 'rangeShiftValue' (or down if this is negative).
// 'rangeShiftValue' is a global RANGE_VALUE_TYPE variable.
DRangeP	RangeShift(DRangeP range);

DRangeP RangeClone(DRangeP);

void	RangePrint(DRangeP r1);

// Prints the values of the range to string buf. Will print a maximum of max_chars.
int	snRangePrint(char *buf, size_t max_chars, DRangeP r1);

int	RangeSize(DRangeP r1);

char	RangeGt(DRangeP, RANGE_VALUE_TYPE value);
char	RangeGtEq(DRangeP, RANGE_VALUE_TYPE value);
char	RangeSt(DRangeP, RANGE_VALUE_TYPE value);
char	RangeStEq(DRangeP, RANGE_VALUE_TYPE value);






#endif
