# error.tcl
#
# tkerror for exmh
#
# Copyright (c) 1993 Xerox Corporation.
# Use and copying of this software and preparation of derivative works based
# upon this software are permitted. Any distribution of this software or
# derivative works must comply with all applicable United States export
# control laws. This software is made available AS IS, and Xerox Corporation
# makes no warranty about the software, its performance or its conformity to
# any specification.

proc Errorlib_Init {tool version {maintainer ""}} {
	global errorlib
	set errorlib(tool) $tool
	set errorlib(version) $version
	if {[string compare $maintainer ""] != 0} {
		set errorlib(maintainer) $maintainer
	}
}

#
# tkerror --
#	This is the handler for background errors that arise
#	from commands bound to keystrokes and menus.  A
#	toplevel message widget is used to display $errorInfo

set ErrorIgnore {
    {^grab failed}
}

proc tkerror { msg } {
    global errorInfo errorlib ErrorIgnore errorCode
    
    Debug_FnArgs tkerror $msg $errorCode
    Debug_Stack tkerror $msg
    
    foreach pat $ErrorIgnore {
	if [regexp $pat $msg] {
	    Exmh_Status $msg
	    return
	}
    }
    set font fixed
    set base ".errorInfo"
    set title "Error Info"
    if [info exists errorInfo] {
	set savedErrorInfo $errorInfo
    } else {
	set savedErrorInfo {no errorInfo}
    }
    # Create a toplevel to contain the error trace back
    if [catch {
	# Choose a unique name
	for {set x 1} {$x<10} {set x [expr $x+1]} {
	    if {! [winfo exists $base-$x]} {
		break
	    }
	}
	set title $title-$x
	set name $base-$x

	set wx [expr ($x%20)*10]
	Widget_Toplevel $name Error $wx $wx
	
	lappend errorlib(errorwins) $name
	bind $name <Destroy> "updateErrorwinInfo $name"

	wm title $name $title
	wm minsize $name 20 5
    
	frame $name.buttons
	pack $name.buttons -side top -fill x
    
	button $name.buttons.quit -text "Dismiss" -command [list destroy $name]
	pack append $name.buttons $name.buttons.quit {left}
	if [info exists errorlib(maintainer)] {
	    button $name.buttons.mailto -text "Mail to $errorlib(maintainer)" -command [list ExmhMailError $name $errorInfo]
	    pack append $name.buttons $name.buttons.mailto {right}
	}
	global widgetText TextType

	message $name.ex -font $font -aspect 1000 -text \
"Please type a few words of explanation before
pressing the Mail to button, or just Dismiss me."
	pack $name.ex -side top -fill x
	text $name.user -font $font -width 60 -bd 2 -relief raised
	$name.user configure -height 5
	$name.user insert end "What happened: "
	$name.user tag add sel 1.0 1.14
	focus $name.user
	pack $name.user -side top -fill both -expand true
	$name.user mark set hlimit 1.0
	set widgetText($name.user,extend) 0
	set widgetText($name.user,geo) {}
	set TextType($name.user) text

	frame $name.msg
	pack $name.msg -side top -fill both -expand true

	text $name.msg.t -font $font -width 60 -bd 2 -relief raised \
		-setgrid true -yscrollcommand [list $name.msg.sy set]
	scrollbar $name.msg.sy -orient vertical -command [list $name.msg.t yview]
	set numLines [llength [split $errorInfo \n]]
	if {$numLines > 20} {
	    set numLines 20
	}
	$name.msg.t configure -height $numLines
	$name.msg.t insert end $errorInfo
	pack $name.msg.sy -side right -fill y
	pack $name.msg.t -side left -fill both -expand true
	set widgetText($name.msg.t,extend) 0
	set widgetText($name.msg.t,geo) {}
	set TextType($name.msg.t) text

	Visibility_Wait $name

    } oops] {
	set msg [concat $msg "($name: " $oops ")" ]
   }

	# print an error message to the error line, or to stderr
    # We only print an error message on stderr if both the followin hold:
    #  - Exmh_Status fails, and
    #  - the main window has been created (i.e. its status != withdrawn)
	set elipses "..."
	set maxlength 30
    if {[string length $msg] > [expr $maxlength + [string length $elipses] + 2]} {
	set msg [concat [string range $msg 0 $maxlength] $elipses]
    }
    if [catch {Exmh_Status "tkerror: $msg" purple} foo] {
    	catch {puts stderr "Exmh_Status fails: $foo"}
    	if {![catch {wm state $path} state]} {
    		if {$state != "withdrawn"} {
				catch {puts stderr "tkrror: $msg"}
				catch {puts stderr "*** TCL Trace ***"}
				catch {puts stderr $savedErrorInfo}
			}
		}
    }
    Debug_FnRes tkerror
}

proc ExmhMailError { w errInfo } {
    global errorlib
    set errorMsgFile [file join [Env_Tmp] errorlibErrorMsg]
    if [catch {open $errorMsgFile w} out] {
	Exmh_Status "Cannot open $errorMsgFile" purple
	return
    }
    if [catch {
	global env tk_version tk_patchLevel tcl_version tcl_patchLevel
	puts $out "To: $errorlib(maintainer)"
	puts $out "Subject: $errorlib(tool) error [lrange $errorlib(version) 1 end]"
	puts $out ""
	set line [$w.user get 1.0 1.end]
	if [regexp {^What happened:} $line] {
	    puts $out [string range $line 14 end]
	} else {
	    puts $out $line
	}
	puts $out [$w.user get 2.0 end]
	puts $out ""
	puts $out "[exec date]"
	if [info exists env(USER)] {
	    puts $out "$env(USER) got an error"
	}
	puts $out "$errorlib(tool) $errorlib(version)"
	if [info exists tk_patchLevel] {
	    set v $tk_patchLevel
	} else {
	    set v $tk_version
	}
	puts $out "TK version $v"
	if [info exists tcl_patchLevel] {
	    set v $tcl_patchLevel
	} else {
	    set v $tcl_version
	}
	puts $out "TCL version $v"
	catch {exec uname -a} uname
	puts $out "$uname"
	puts $out ""
	puts $out $errInfo
	close $out
    } msg] {
	Exmh_Status "$errorMsgFile $msg" purple
	return
    }
    if [catch {
	# exec send $errorMsgFile
	exec /usr/lib/sendmail $errorlib(maintainer) < $errorMsgFile
    } msg] {
	Exmh_Status "Send error: $msg" purple
	return
    } else {
	Exmh_Status "Mailed report to $errorlib(maintainer)"
	destroy $w
    }
    catch {exec rm -f $errorMsgFile}
}
proc bgerror [info args tkerror] [info body tkerror]

# We set bgerror, because we use fileevent to
# handle the input, and fileevent runs the input-handler
# as a background command
proc mybgerror {msg} {
	global errorInfo errorCode
	catch {puts stderr "Fatal InternalError ($errorCode): $msg"}
	set lines [split $errorInfo \n]
	catch {puts stderr  "Stack Begin"}
	foreach l $lines {
		catch {puts stderr  "Stack: $l"}
	}
	catch {puts stderr  "Stack End"}
	if {([string compare $errorCode NONE] == 0) &&
	    ([string compare $errorInfo ""] == 0)} {
	}
	# exit
}



proc updateErrorwinInfo { name } {
	global errorlib noMoreErrorWinsAlive
    Debug_FnArgs updateErrorwinInfo $name
	if {[info exists errorlib(errorwins)]} {
		set found [lsearch -exact $errorlib(errorwins) $name]
		Debug_Msg updateErrorwinInfo "$name $found [llength $errorlib(errorwins)]"
		if {$found != -1} {
			set errorlib(errorwins) [lreplace $errorlib(errorwins) $found $found]
		}
		Debug_Msg updateErrorwinInfo "$name [llength $errorlib(errorwins)]"
		if {[llength $errorlib(errorwins)] == 0} {
			set noMoreErrorWinsAlive 1
		}
	} else {
		set noMoreErrorWinsAlive 1
	}
    Debug_FnRes updateErrorwinInfo
}
proc waitErrorWindows {} {
	global errorlib noMoreErrorWinsAlive
    Debug_FnArgs waitErrorWindows
	if {[info exists errorlib(errorwins)] && ([llength $errorlib(errorwins)] > 0)} {
		tkwait variable noMoreErrorWinsAlive
	}
    Debug_FnRes waitErrorWindows
}

proc Exmh_Debug {fn args} {
	catch {puts stderr "Exmh_Debug $fn $args"}
}

proc Exmh_Error { msg } {
# catch {puts stderr "Exmh_Error(start): $msg" }
    Debug_FnArgs Exmh_Error $msg
    Debug_Stack Exmh_Error $msg
    global errorInfo
    set errorInfo {}
    set level [info level]
    for {set l 0} {$l <= $level} {incr l} {
	append errorInfo [info level $l]
	append errorInfo \n
    }
    tkerror $msg
# catch {puts stderr "Exmh_Error(end): $msg" }
}

proc Visibility_Wait {win} {
    catch {tkwait visibility $win}
}
proc Exmh_Status {string { level normal } } {
	if {[catch {Exmh_do_Status $string $level} foo]} {
		catch {puts stderr $string}
	}
}
proc Exmh_do_Status {string { level normal } } {
	Debug_FnArgs Exmh_do_Status $string $level
	set wl [WcFullNameToWidget "." "*error_line"]
	set awl [WcFullNameToWidget "." "*adt_error_line"]
	set_label $wl $string
	set_label $awl $string
}
