# =========================================================================
#
# main loop
#
# precondition:
#  we have already computed a menu, based on the menus of all primers,
#  which is in menuinfo(inputs),  menuinfo(outputs).
#  menuinfo(inconsistent) = 1 if both input and output menu are empty
#
#

proc main_loop {} {
	global maininfo global menuinfo specinfo sidinfo user_spawn_id
	if {$maininfo(debug) > 1} {
		send_error "DEBUG: in mainloop, waiting for commands...\n"
	}
	while {[main_go_on]} {
		expect_user {
			-i $user_spawn_id eof {
				we_stop
			}
			-re "^C_GETCONFIG\[ \t]*\r*\n" {
				send_user "A_ERROR UnknownCommand: C_GETCONFIG\n"
			}
			-re "^C_EOT\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_EOT\n"
				}
				send_menu_verdict
				set eot 0
				if {$menuinfo(inconsistent)} {
					set eot 1
				} else {
					foreach p $specinfo(primerids) {
						set result [spec_test_eot $p]
						switch -- $result \
							no {
							} yes {
								set eot 1
							} error {
								set eot 1
							}
					}
				}
				if {$eot} {
					send_user "A_EOT_YES\n"
				} else {
					send_user "A_EOT_NO\n"
				}
			}
			-re "^C_IOKIND\[ \t]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set iokind_str [field_value $fields iokind]
				set sug $iokind_str
				set channel_str [field_value $fields channel]
				set chan $channel_str
				if {$maininfo(debug) > 1} {
					send_user "DEBUG: handling C_IOKIND\n"
				}
				send_menu_verdict
				if {([llength $menuinfo(inputs)] == 0)} {
					send_user "A_IOKIND\tiokind=output\tchannel=out\n"
				} elseif {([llength $menuinfo(outputs)] == 0)} {
					send_user "A_IOKIND\tiokind=input\tchannel=in\n"
				} elseif {($sug != "")} {
					set pfx "A_IOKIND\tiokind=$sug"
					if {[string compare $chan ""] != 0} {
						append pfx "\tchannel=$chan"
					}
					send_user "$pfx\n"
				} elseif {[llength $specinfo(spectestids)] > 0} {
					set result [spec_get_iokind [lindex $specinfo(spectestids) 0] ""]
					switch -- $result \
						input {
							send_user "A_IOKIND\tiokind=input\tchannel=in\n"
						} output {
							send_user "A_IOKIND\tiokind=output\tchannel=out\n"
						} error {
							send_user "A_ERROR spec returned error\n"
							we_stop
						}
				} else {
					# do something random
					send_user "A_IOKIND\tiokind=input\tchannel=in\n"
				}
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handled C_IOKIND\n"
				}
			}
			-re "^C_INPUTS\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_INPUTS\n"
				}
				send_menu_verdict
				send_user "A_INPUTS_START\n"
				foreach i [lsort $menuinfo(inputs) ] {
					set pfx  "A_EVENT\tevent=$i\tchannel=in"
					foreach ap  [attr_table_get_attrs inputs $i] {
						append pfx "\t"
						append pfx [lindex $ap 0]
						append pfx "="
						append pfx [lindex $ap 1]
					}
					send_user "$pfx\n"
				}
				send_user "A_INPUTS_END\n"
			}
			-re "^C_OUTPUTS\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_OUTPUTS\n"
				}
				send_menu_verdict
				send_user "A_OUTPUTS_START\n"
				foreach i [lsort $menuinfo(outputs) ] {
					set pfx "A_EVENT\tevent=$i\tchannel=out"
					foreach ap  [attr_table_get_attrs outputs $i] {
						append pfx "\t"
						append pfx [lindex $ap 0]
						append pfx "="
						append pfx [lindex $ap 1]
					}
					send_user "$pfx\n"
				}
				send_user "A_OUTPUTS_END\n"
			}
			-re "^C_GETINPUT\[ \t]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set event_str [field_value $fields event]
				set sug $event_str
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_GETINPUT $fields\n"
				}
				send_menu_verdict
				set result [spec_choose_from_menu $sug]
				set input [lindex $result 1]
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: choosen input: $input\n"
				}
				foreach p $specinfo(primerids) {
					set result [spec_get_input $p $input]
					switch -- [lindex $result 0] \
						input_ok {
							# check that primer result is identical to
							# result of [spec_choose_from_menu $sug]
							set primerinput [lindex $result 1]
							if {![event_match $input $primerinput] &&
							    ![event_eq $input $primerinput]} {
								send_user "A_WARNING input mismatch for $p: ($input)($primerinput)\n"
							}
						} input_error {
							set err "inconsistency when requesting input from $p\n"
							send_user "A_ERROR $err"
							fatal $err
						} error {
							set err "inconsistency when requesting input from $p\n"
							send_user "A_ERROR $err"
							fatal $err
						}
				}
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: A_GETINPUT_OK event=$input\tchannel=in\n"
				}
				send_user "A_GETINPUT_OK event=$input\tchannel=in\n"
			}
			-re "^C_INPUT\[ \t]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set event_str [field_value $fields event]
				set sug $event_str
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_INPUT $fields\n"
				}
				send_menu_verdict
				set result [spec_choose_from_menu $sug]
				if {$maininfo(debug) > 1} {
					send_error "DEBUG:    handling C_INPUT $fields spec_choose_from_menu resul=($result)\n"
				}
				set input [lindex $result 1]
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: choosen input: $input\n"
				}
				foreach p $specinfo(primerids) {
					set result [spec_do_input $p $input]
					switch -- [lindex $result 0] \
						input_ok {
							# check that primer result is identical to
							# result of [spec_choose_from_menu $sug]
							set primerinput [lindex $result 1]
							if {![event_match $input $primerinput] &&
							    ![event_eq $input $primerinput]} {
								send_user "A_WARNING input mismatch for $p: ($input)($primerinput)\n"
							}
						} input_error {
							set err "inconsistency when requesting input from $p\n"
							send_user "A_ERROR $err"
							fatal $err
						} error {
							set err "inconsistency when requesting input from $p\n"
							send_user "A_ERROR $err"
							fatal $err
						}
				}
				update_menu
				send_user "A_INPUT_OK event=$input\tchannel=in\n"
			}
			-re "^C_OUTPUT\[ \t]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set event_str [field_value $fields event]
				set suspension_str [field_value $fields suspension]
				set channel_str [field_value $fields channel]
				set output $event_str
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_OUTPUT $fields\n"
				}
				send_menu_verdict
				set all_ok 1
				### HACK: we unset specresult for the HACK below
				catch {unset specresult}
				# first, see whether or not the event-to-check is the delta-event
				set is_delta_event 0
				if {([string compare $suspension_str ""] != 0)} {
					if {$suspension_str} {
						set is_delta_event 1
					} else {
						set is_delta_event 0
					}
				} elseif {[info exists specinfo(label,delta,this)]} {
					if {[event_eq $output $specinfo(label,delta,this)]} {
						set is_delta_event 1
					}
				} elseif {[string match $output ""]} {
					set is_delta_event 1
				}
				set failed 0
				foreach p $specinfo(spectestids) {
					if {$is_delta_event} {
						if {[info exists specinfo(label,delta,$p)]} {
							set spec_output $specinfo(label,delta,$p)
						} else {
							# default value
							set spec_output Delta
						}
					} else {
						set spec_output $output
					}
					set specresult [spec_check_output $p $spec_output $is_delta_event $channel_str]
					switch -- [lindex $specresult 0] \
						output_ok {
							# Ok
						} output_error {
							set all_ok 0
							set failed 1
						}
				}
				foreach p $specinfo(guideids) {
					if {$is_delta_event} {
						if {[info exists specinfo(label,delta,$p)]} {
							set guide_output $specinfo(label,delta,$p)
						} else {
							# default value
							set guide_output "guidance ! DELTA"
						}
					} else {
						set guide_output $output
					}
					# we use algorithm ``traces'' for guide, so it will know
					# about suspension
					# (otherwise: use ``0'' instead of  $is_delta_event)
					set result [spec_check_output $p $guide_output $is_delta_event $channel_str]
					switch -- [lindex $result 0] \
						output_ok {
							# Ok
							if {$failed} {
								# check whether we reached end-of-trace
								if {![menu_contains_hit $p]} {
									set menuinfo(verdict) MISS
									set menuinfo(verdict,seen) 1
								}
							}
						} output_error {
							send_user "A_GUIDANCE_VERDICT MISS $p\n"
							# send_user "A_GUIDANCE_VERDICT INCONCLUSIVE $p\n"
							set all_ok 0
						}
				}
				### HACK
				if {![info exists specresult]} {
					if {$all_ok} {
						set specresult output_ok
					} else {
						set specresult output_error
					}
				}
				switch -- [lindex $specresult 0] \
					output_ok {
						if {$all_ok} {
							update_menu
						}
						send_user "A_OUTPUT_OK\tevent=$output\tsuspension=$is_delta_event\tchannel=$channel_str\n"
					} output_error {
						send_menu_verdict
						send_user "A_OUTPUT_ERROR\n"
					}
			}
			-re "^C_STATE\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_STATE\n"
				}
				send_menu_verdict
				send_user "A_STATE_START\n"
				foreach p $specinfo(primerids) {
					set result [spec_get_state $p]
					switch -- [lindex $result 0] \
						state {
							send_user "== Start State for $p:\n"
							set state [lindex $result 1]
							foreach e $state {
								send_user "$e\n"
							}
							send_user "== End State for $p ==\n"
						} error {
						}
				}
				send_user "A_STATE_END\n"
			}
			-re "^C_STATS\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_STATS\n"
				}
				send_menu_verdict
				send_user "A_STATS_START\n"
				foreach p $specinfo(primerids) {
					set result [spec_get_stats $p]
					set r $specinfo(role,$p)
					switch -- [lindex $result 0] \
						stats {
							set stats [lindex $result 1]
							foreach e $stats {
								send_user "A_STATS id $p role $r $e\n"
							}
						} error {
						}
				}
				send_user "A_STATS_END\n"
			}
			-re "^C_STATEID\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_STATEID\n"
				}
				send_menu_verdict
				send_user "A_STATEID_START\n"
				foreach p $specinfo(primerids) {
					if {![info exists maininfo(nostateid,$p)]} {
						set result [spec_get_stateid $p]
						set r $specinfo(role,$p)
						switch -- [lindex $result 0] \
							stateid {
								set stats [lindex $result 1]
								foreach e $stats {
									set fields [fields_split $e]
									set identity [field_value $fields identity]
									set super [field_value $fields super]
									set init [field_value $fields init]
									set trans [field_value $fields trans]
									set identity [concat id $p role $r $identity]
									set newfields [fields_join \
										[list \
											[list identity [join $identity " "]] \
											[list super $super] \
											[list init $init ] \
											[list trans $trans ]\
										] \
									]
									send_user "A_STATEID $newfields\n"
								}
							} error {
								set maininfo(nostateid,$p) 1
							}
					}
				}
				send_user "A_STATEID_END\n"
			}
			-re "^C_GOTO\[ \t]*(\[^\r\n]*)\r*\n" {
				set fields [fields_split $expect_out(1,string)]
				set tuples [field_value $fields super]
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_GOTO $fields\n"
				}
				set result ""
				set mode normal
				set super ""
				set id ""
				set role ""
				while {[string compare $result ""] == 0 && [llength $tuples] > 0} {
					set word [lindex $tuples 0]
					set tuples [lrange $tuples 1 end]
					switch -- $mode \
						  normal {
							if {[string compare $super ""] != 0 && [string compare $id ""] != 0 && [string compare $role ""] != 0 } {
								set result [ spec_goto $super $id $role]
								switch -- [lindex $result 0] \
								goto {
								} error {
								}
								set super ""
								set id ""
								set role ""
								set result ""
							}
							switch -- $word \
							  super {
								set mode super
							} id {
								set mode id
							} role {
								set mode role
							} default {
								set super $word
							}
						} super {
							set super $word
							set mode normal
						} id {
							set id $word
							set mode normal
						} role {
							set role $word
							set mode normal
						}
				}
				if {[string compare $id ""] == 0} {
					send_user "A_ERROR InternalError goto id unset; please provide super state information and id for each connected spec guide or test\n"
				} else {
					if {[string compare $result ""] == 0 && [string compare $super ""] != 0} {
						set result [spec_goto $super $id $role]
						switch -- [lindex $result 0] \
							goto {
							} error {
							}
							set super ""
							set id ""
							set role ""
							set result ""
					}
					if {[string compare $result ""] != 0} {
						send_user "A_GOTO_ERROR $result\n"
					} else {
						update_menu
						send_menu_verdict
						send_user "A_GOTO_OK\n"
					}
				}
			}
			-re "^C_GUIDANCE_INFO\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_GUIDANCE_INFO\n"
				}
				send_user "A_GUIDANCE_INFO_START\n"
				foreach p $specinfo(primerids) {
					send_user "A_GUIDANCE_INFO $p\t$specinfo(role,$p)\n"
				}
				send_user "A_GUIDANCE_INFO_END\n"
			
			}
			-re "^C_GUIDANCE_QUIT\[ \t]*(\[^\r\n]*)\r*\n" {
				set guideids $expect_out(1,string)
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_GUIDANCE_QUIT $guideids\n"
				}
				set quitted {}
				foreach p $specinfo(primerids) {
					if {[lsearch -exact $guideids $p] >= 0} {
						set result [spec_quit $p]
						switch -- [lindex $result 0] \
							quit {
								lappend quitted $p
							} error {
							}
					}
				}
				# we want to update the menu BEFORE sending
				# the A_GUIDANCE_QUIT to make sure that all
				# A_GUIDANCE_VERDICT output that _may_ be generated by
				# update_menu will appear before the A_GUIDANCE_QUIT
				if {[llength $specinfo(primerids)] > 0} {
					update_menu
				}
				send_user "A_GUIDANCE_QUIT $quitted\n"
				if {[llength $specinfo(primerids)] <= 0} {
					send_user "A_ERROR Quitted all primers\n"
					send_user "A_QUIT\n"
					we_stop
				}
			}
			-re "^C_QUIT\[ \t]*\r*\n" {
				if {$maininfo(debug) > 1} {
					send_error "DEBUG: handling C_QUIT\n"
				}
				send_menu_verdict
				foreach p $specinfo(primerids) {
					send -i  $sidinfo(spec,$p) "C_QUIT\n"
				}
				send_user "A_QUIT\n"
				we_stop
			}
			-re "^(\[^\r\n]*)\r*\n" {
				set cmd $expect_out(1,string)
				send_user "A_ERROR UnknownCommand: $cmd\n"
			}
		}
	}
}

# =========================================================================
