proc event_table_clear {} {
	global event_table table_info attr_table attr_info
	catch {unset event_table}
	catch {unset table_info}
	catch {unset attr_table}
	catch {unset attr_info}
}

# we insert: channel primerid event attributes
# where attributes is a list of name-value pairs
proc event_table_insert { channel id ev {attr {}}} {
	global maininfo event_table table_info
	if {$maininfo(debug) > 3} {
		send_error "DEBUG: event_table_insert:  ($channel)($id)($ev)($attr)\n"
	}
	set ev [event_canonical $ev]
	set table_info(spec,$id) 1
	if {[string first * $ev] >= 0} {
		event_table_add_pattern $channel $id $ev $attr
	} else {
		event_table_do_insert $channel $id $ev $attr
	}
}
proc event_table_add_pattern { channel id pat attr } {
	global maininfo event_table table_info
	if {$maininfo(debug) > 3} {
		send_error "DEBUG: event_table_add_pattern:  ($channel)($id)($pat)($attr)\n"
	}
	set evs {}
	foreach name [array names event_table] {
		if {$maininfo(debug) > 3} {
			send_error "DEBUG: event_table_add_pattern:  ($channel)($name)\n"
		}
		if {[regexp "^$channel,(.*)$" $name ignore event]} {
			if {[string match $pat $event]} {
				lappend evs $event
				# send_error "DEBUG: event_table_add_pattern matched pat=($pat) event=($event)\n"
				event_table_do_insert $channel $id $event {} ;# to increment
				event_table_add_mapping $channel $id $event $pat
			} else {
				# send_error "DEBUG: event_table_add_pattern nonmatch pat=($pat) event=($event)\n"
			}
		}
	}
	set norm_attr [attr_normalize $attr [llength $evs]]
	foreach ev $evs {
		attr_table_do_insert $channel $id $ev $norm_attr 
	}
}

proc attr_normalize {attr n} {
	set normalized {}
	foreach a $attr {
		set nm [lindex $a 0]
		set val [lindex $a 1]
		if {[lsearch $nm [list weight]] >= 0} {
			set val [expr ($val * 1.0)/ $n]
		}
		lappend normalized [list $nm $val]
	}
	return $normalized
}

proc event_table_add_mapping { channel id ev pat } {
	global maininfo mapping_table table_info
	set ev [event_canonical $ev]
	if {![info exists table_info(hasmap,$id)]} {
		set table_info(hasmap,$id) 1
	}
	set mapping_table($id,$ev) [event_canonical $pat]
}
proc event_table_lookup_mapping { channel id orgev } {
	global maininfo mapping_table table_info
	set ev [event_canonical $orgev]
	if {[info exists table_info(hasmap,$id)]} {
		if {[info exists mapping_table($id,$ev)]} {
			return $mapping_table($id,$ev)
		} else {
			foreach name [array names mapping_table] {
				set pat $mapping_table($name)
				if {[string match $pat $ev]} {
					return $ev
				} else {
					# send_user "A_WARNING event_table_lookup_mapping tried name=($name) pat=($pat) ev=($ev)\n"
				}
			}
			send_user "A_WARNING event_table_lookup_mapping (internal error?): no $id mapping for: $orgev\n"
		}
	}
	return $ev
}

proc event_table_register { channel id } {
	global table_info
	if {![info exists table_info(primers,$channel,$id)]} {
		set table_info(primers,$channel,$id) 1
		lappend table_info(primernames,$channel) $id
		if {[info exists table_info(primers,$channel)]} {
			incr table_info(primers,$channel)
		} else {
			set table_info(primers,$channel) 1
		}
	}
}

proc event_table_primers { channel } {
	global table_info
	if {[info exists table_info(primernames,$channel)]} {
		return $table_info(primernames,$channel)
	} else {
		return {}
	}
}


proc event_table_do_insert { channel id ev attrpairs } {
	global maininfo event_table table_info
	if {$maininfo(debug) > 3} {
		send_error "DEBUG: event_table_do_insert:  ($channel)($id)($ev)($attrpairs)\n"
	}
	if {[info exists event_table($channel,$ev)]} {
		incr event_table($channel,$ev)
	} else {
		set event_table($channel,$ev) 1
	}
	event_table_register $channel $id
	attr_table_do_insert $channel $id $ev $attrpairs 
}

proc attr_table_do_insert { channel id ev attrpairs } {
	global attr_table attr_info
	foreach ap $attrpairs {
		set key [lindex $ap 0]
		set val [lindex $ap 1]
		if {![info exists attr_table($channel,$ev,$key)]} {
			set attr_table($channel,$ev,$key) $val
		} elseif {[string compare $attr_table($channel,$ev,$key) $val] != 0} {
			send_error "Warning: event_table_do_insert: differing values for attr $key of event \"$ev\": \"$attr_table($channel,$ev,$key)\" ne \"$val\"\n"
		}
		if {![info exists attr_info($channel,$ev,$key)]} {
			set attr_info($channel,$ev,$key) 1
			lappend attr_info($channel,$ev) $key
		}
	}
}

proc attr_table_get_attrs { channel ev } {
	global attr_table attr_info
	set res {}
	if {[info exists attr_info($channel,$ev)]} {
		foreach key $attr_info($channel,$ev) {
			lappend res [list $key $attr_table($channel,$ev,$key)]
		}
	}
	return $res
}

proc event_table_intersect { channel } {
	global maininfo event_table table_info
	set events {}
	set nevents 0
	if {[info exists table_info(primers,$channel)]} {
		set nr_of_primers $table_info(primers,$channel)
		foreach name [array names event_table] {
			if {$maininfo(debug) > 3} {
				send_error "DEBUG: event_table_intersect:  ($channel)($name)(== $event_table($name); nr_of_primers=$nr_of_primers)\n"
			}
			if {[regexp "^$channel,(.*)$" $name ignore event]} {
				incr nevents
				if {$event_table($name) == $nr_of_primers} {
					lappend events $event
				} elseif {$event_table($name) > $nr_of_primers} {
					send_user "A_WARNING event_table_intersect (internal error?): (event_table(name) == $event_table($name)) > (nr_of_primers == $nr_of_primers) for event: $event\n"
				}
			}
		}
	}
	return [list $nr_of_primers $nevents $events]
}
