# =========================================================================
#
# Initialisation / config file processing
#

proc init_variables {} {
	global gateinfo specinfo maininfo configinfo cmdarginfo

	# General Configuration values (that can not be configured by the user)
	set maininfo(timeout) 10

	# defaults for user config values

	set maininfo(debug) 0
	set configinfo(DEBUG) $maininfo(debug)

	set maininfo(nrofstepsdone) 0
	set maininfo(stopnow) 0
	set maininfo(chooseInputs) 0
	set configinfo(CHOOSEINPUTS) $maininfo(chooseInputs)

	if {[info exists cmdarginfo(seed)]} {
		set specinfo(seed) $cmdarginfo(seed)
	} else {
		set specinfo(seed) 4
		# set specinfo(seed) 5
	}
	set configinfo(SEED) $specinfo(seed)

	if {[info exists cmdarginfo(algo)]} {
		set specinfo(algo) $cmdarginfo(algo)
		set configinfo(ALGO) $specinfo(algo)
	} else {
		set specinfo(algo) ""
	}

	set specinfo(timeout) -1 ;# effectively disables timeout for spec!!!
	set configinfo(SPECTIMEOUT) $specinfo(timeout)
	
	if {[info exists cmdarginfo(depth)]} {
		set maininfo(nrofstepsmax)  $cmdarginfo(depth)
	} else {
		set maininfo(nrofstepsmax) 0
	}
	set configinfo(MAXSTEPS) $maininfo(nrofstepsmax)
	
	set specinfo(primerids) {}	;# list of _all_ primer ids (spec, test, guide)
	set specinfo(specids) {}	;# list of spec primer ids
	set specinfo(testids) {}	;# list of test primer ids
	set specinfo(guideids) {}	;# list of guide primer ids

	if {![info exists gateinfo(igates)]} {
		set gateinfo(igates) {}
	}
	if {![info exists gateinfo(ogates)]} {
		set gateinfo(ogates) {}
	}
	
	# reserve the id 'this' for us, i.e. for the current tool
	# currently we simply map it on the empty string, the thing
	# we map to will not be used, anyway...
	set specinfo(config,this) ""
	set configinfo(THIS\ this) $specinfo(config,this)

	# set initial delta label for us, i.e. for the current tool,
	# i.e. the delta label that will be used between us and our driver
	# Note that this (from now on) will be used in a symmetric way, i.e.
	# both in the menu's that we send to the Driver, and also in the
	# (input, output) events (or suggestions) that the Driver sends
	# to us.
	if {[info exists cmdarginfo(delta)]} {
		set specinfo(label,delta,this) $cmdarginfo(delta)
	} else {
		set specinfo(label,delta,this) "Delta"
	}
	set configinfo(LABEL-DELTA\ this) $specinfo(label,delta,this)
	
}

proc process_config_line_words { file line infolist } {
	global gateinfo specinfo maininfo configinfo

	set iomap(INPUT) input
	set iomap(OUTPUT) output

	foreach e $infolist {
		set c [string toupper [lindex $e 0]]
		if {[string compare "SPECTIMEOUT" $c] == 0} {
			set specinfo(timeout) [lindex $e 1]
			set configinfo($c) $specinfo(timeout)
		} elseif {([string compare "CHOOSEINPUTS" $c] == 0)} {
			set maininfo(chooseInputs)  [lindex $e 1]
			set configinfo($c) $maininfo(chooseInputs)
		} elseif {([string compare "SEED" $c] == 0)} {
			set specinfo(seed)  [lindex $e 1]
			set configinfo($c) $specinfo(seed)
		} elseif {([string compare "PARTFILE" $c] == 0)} {
			set maininfo(partfile)  [lindex $e 1]
			set configinfo($c) $maininfo(partfile)
		} elseif {([string compare "DEBUG" $c] == 0)} {
			set maininfo(debug)  [lindex $e 1]
			set configinfo($c) $maininfo(debug)
		} elseif {([string compare "SPEC" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_defined $id $file $line
			set specinfo(config,$id) [lrange $c 2 end]
			if {![catch {glob [lindex $e 2]} foo]} {
			} elseif {![catch {glob [file join $maininfo(libexecdir) [lindex $e 2]]} foo]} {
			} elseif {![catch {glob [file join $maininfo(bindir) [lindex $e 2]]} foo]} {
			} else {
				fatal "config $file:$line: Could not find specified $c program: [lindex $e 2]";
			}
			set specinfo(program,$id) $foo
			set specinfo(role,$id) spec
			set configinfo($c\ $id) $specinfo(program,$id)
			lappend specinfo(specids) $id
			lappend specinfo(primerids) $id
		} elseif {([string compare "SPECFLAGS" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_undefined $id $file $line
			lappend specinfo(program,flags,$id) [join [lrange $e 2 end] " "]
			lappend configinfo($c\ $id) $specinfo(program,flags,$id)
		} elseif {([string compare "TEST" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_defined $id $file $line
			set specinfo(config,$id) [lrange $c 2 end]
			if {![catch {glob [lindex $e 2]} foo]} {
			} else {
				fatal "config $file:$line: Could not find specified $c program: [lindex $e 2]";
			}
			set specinfo(program,$id) $foo
			set specinfo(role,$id) test
			set configinfo($c\ $id) $specinfo(program,$id)
			lappend specinfo(testids) $id
			lappend specinfo(primerids) $id
		} elseif {([string compare "GUIDE" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_defined $id $file $line
			set specinfo(config,$id) [lrange $c 2 end]
			if {![catch {glob [lindex $e 2]} foo]} {
			} else {
				fatal "config $file:$line: Could not find specified $c program: [lindex $e 2]";
			}
			set specinfo(program,$id) $foo
			set specinfo(role,$id) guide
			set configinfo($c\ $id) $specinfo(program,$id)
			lappend specinfo(guideids) $id
			lappend specinfo(primerids) $id
		} elseif {([string compare "RUNDIR" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_undefined $id $file $line
			set specinfo(program,run,dir,$id) [lindex $e 2]
			## we should check here
			set configinfo($c\ $id) $specinfo(program,run,dir,$id)
		} elseif {([string compare "LABEL-DELTA" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_undefined $id $file $line
			set specinfo(label,delta,$id) [lrange $e 2 end]
			set configinfo($c\ $id) $specinfo(label,delta,$id)
		} elseif {([string compare "LABEL-HIT" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_undefined $id $file $line
			set specinfo(label,hit,$id) [lrange $e 2 end]
			set configinfo($c\ $id) $specinfo(label,hit,$id)
		} elseif {([string compare "LABEL-MISS" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_undefined $id $file $line
			set specinfo(label,miss,$id) [lrange $e 2 end]
			set configinfo($c\ $id) $specinfo(label,miss,$id)
		} elseif {([string compare "LABEL-PASS" $c] == 0)} {
			set id [lindex $e 1]
			fatal_if_id_undefined $id $file $line
			set specinfo(label,pass,$id) [lrange $e 2 end]
			set configinfo($c\ $id) $specinfo(label,pass,$id)
		} else {
			send_user "unexpected: $e\n"
			lappend configinfo(unexpected) $e
		}
	}

	set specinfo(spectestids) [concat $specinfo(specids) $specinfo(testids) ]

}

proc fatal_if_id_defined {id file line} {
	global specinfo
	if {[string compare "this" $id] == 0} {
		fatal "config $file:$line: id \"this\" is reserved for current tool\n"
	}
	if {[info exists specinfo(config,$id)]} {
		fatal "config $file:$line: Spec/Guide/Test id occurs more than once: $id\n"
	}
}

proc fatal_if_id_undefined {id file line} {
	global specinfo
	if {![info exists specinfo(config,$id)]} {
		fatal "config $file:$line: unknown Spec/Guide/Test id (should be defined before use): $id\n"
	}
}

# placeholder
proc generate_primer_commands {} {}

proc generate_primer_command {id} {
	global specinfo gateinfo

		if {[info exists specinfo(program,run,dir,$id)]} {
			set cdcmd "cd $specinfo(program,run,dir,$id)"
			set program $specinfo(program,$id)
		} else {
			set cdcmd "cd [file dirname $specinfo(program,$id)]"
			set program [file join . [file tail $specinfo(program,$id)]]
		}
		set cmd_string $program
		if {[llength $specinfo(seed)] > 0} {
			append cmd_string " -s $specinfo(seed)"
		}
		if {[llength $gateinfo(igates)] > 0} {
			append cmd_string " -i $gateinfo(igates)"
		}
		if {[llength $gateinfo(ogates)] > 0} {
			append cmd_string " -o $gateinfo(ogates)"
		}
		if {[llength $specinfo(algo)] > 0} {
			append cmd_string " -S $specinfo(algo)"
		}
		set delta_label [get_delta_label $id]
		if {[llength $delta_label] > 0} {
			append cmd_string " -d \"$delta_label\""
		}
		if {[info exists specinfo(program,flags,$id)]} {
			append cmd_string " $specinfo(program,flags,$id)"
		}
		set specinfo(cmd,$id) [list spawn -noecho "/bin/sh" -c "stty -echo; $cdcmd; exec 2> $specinfo(pty,$id,stderr); $cmd_string"]
		# puts stderr "generate_primer_commands ($id)($specinfo(cmd,$id))"

}

proc spec_id {name} {
	global specinfo
	if {[info exists specinfo($name)]} {	
		return $specinfo($name)
	} else {
		fatal "Internal error: unknown Spec/Guide: $name\n"
	}
}

proc initialise_random_seed {} {
	global specinfo
	expr srand($specinfo(seed))
}

proc get_delta_label {name} {
	global specinfo
	if {[info exists specinfo(label,delta,$name)]} {
		return $specinfo(label,delta,$name)
	} else {
		return ""
	}
}

# =========================================================================
