# =========================================================================
# $Id: next.tcl,v 1.1.2.15 2008/04/11 11:54:06 belinfan Exp $
# =========================================================================

proc primer_next {kind ev chan upreds} {
	global primer
	set evs {}
	set errmsg ""
	set matches [try_predicates [try_instantiate $kind $ev $chan]]
#	if {[llength $matches] == 0 && [string compare $upreds ""] != 0} {
#		debugMsg "primer_next no matches"
#		set matches [list [list $ev ""]]
#	}
	
	foreach m $matches {
		set sm_ev [lindex $m 0]
		set sm_p [lindex $m 1]
		if {[string compare $upreds ""] != 0} {
			set u_p [subst_variables $kind $chan $sm_ev $upreds]
			set preds "$u_p $sm_p"
		} else {
			set preds $sm_p
		}
		set sm_id [expansion_identical $sm_ev]
		set sm_id2 [mr_identical $sm_ev]
		debugMsg "m=($m) sm_ev=($sm_ev) id=($sm_id:$sm_id2) sm_p=($sm_p) upreds=($upreds) preds=($preds)"
		if {[string compare $preds ""] != 0} {
			set instantiate_res [cache_explorer_instantiate $sm_ev $sm_id2 $preds]
			debugMsg "instantiate_res($sm_ev=$sm_id:$sm_id2)($preds)=$instantiate_res"
			if {[lindex $instantiate_res 0]} {
				set new_ev [lindex $instantiate_res 1]
				set e_nr [lindex $new_ev 0]
				set visible [lindex $new_ev 1]
				set preds_solved  [lindex $new_ev 2]
				set repr [lindex $new_ev 3]
				set preds  [lindex $new_ev 4]
				set freevars  [lindex $new_ev 5]
				set identical [lindex $new_ev 6]
				if {! $preds_solved} {
			  		warningMsg "Predicates Unsolved for $kind $e_nr: $preds"
				}
				# we no longer insert the result of instantiation,
				# because it is implicitly already included in the
				# uninstantiated symbolic event, and because
				# timed instantiation results may be too late,
				# and if so,we have to retry. when we retry,
				# we first select the event to do. if we would
				# insert the instantiation results, we might
				# once more choose an older instantiation result -
				# one that we actually want to avoid
				# mr_insert $kind $chan $new_ev $repr
				set newSmap($e_nr) 1
				lappend evs $new_ev
			} else {
			  	# warningMsg "Instantiation fails for $kind: $ev \[$preds\]: [lindex $instantiate_res 1]"
			  	append errmsg " Instantiation fails for $kind: $ev \[$preds\]: [lindex $instantiate_res 1]."
			}
		} else {
			set newSmap([lindex $sm_ev 0]) 1
			lappend evs [mr_event $kind $chan $sm_ev]
		}
	}
	set newS [array names newSmap]
	debugMsg "primer_next $kind $chan ($ev) ($upreds): [list $evs $newS]"
	set l [llength $newS]
	if {($l > 0) && [string compare $primer(algorithm) sim] == 0} {
		debugMsg "The number of new states is $l"
		set r [random $l]
		set s [lindex $newS $r]
		set newS [list $s]
		set evs [list [lindex $evs $r]]
		debugMsg "We randomly choose state $r: $s"
	}
	return [list $evs $newS $errmsg]
}

# ==========================================

proc try_instantiate {kind ev chan} {
	set result {}
	set t [event2term $ev]
	switch -- $kind {
	  inputsuspension -
	  outputsuspension -
	  input -
	  output {
		foreach el [mr_list [list $kind ] [list $chan]] {
			set el_kind [lindex $el 0]
			set el_chan [lindex $el 1]
			debugMsg "try_instantiate-el ($kind): $el"
			set mev [lindex $el 2]
			set mt [event2term $mev]
			set unification [unify $mt $t]
			debugMsg "try_instantiate-unification ($kind)($el): ($unification)"
			if {[lindex $unification 0]} {
				set smile_events [lindex $el 3]
				bindings_prepare [lindex $unification 1]
				set tmp_result {}
				foreach sm_el $smile_events {
					debugMsg "try_instantiate-sm_el ($sm_el)"
					set sm_bindings [bindings_update $kind $chan $sm_el]
					lappend tmp_result [list $sm_el $sm_bindings]
				}
				lappend result [list $el [lindex $unification 1] $tmp_result]
				bindings_clear
			}
		}
	  }
	}
	debugMsg "try_instantiate ($kind) ($ev) ($chan): ($result)"
	return $result
}

proc try_predicates {trylist} {
	set result {}
	foreach m $trylist {
		set uni_smile_events [lindex $m 2]
		debugMsg "try_predicates-el: ($m) ($uni_smile_events)"
		foreach e $uni_smile_events {
			lappend result [make_predicates	$e]
		}
	}
	debugMsg "try_predicates ($trylist): ($result)"
	return $result
}

proc make_predicates {unif} {
	set result ""
	set ev [lindex $unif 0]
	set bindings [lindex $unif 1]
	foreach b $bindings {
		set v [lindex $b 0]
		set e [lindex $b 1] ;# single expression (single experiment-option)
		set t [list $e] ;# encapsulate single experiment in a term
		set s [term2event $t]
		set p "$v = $s; "
		append result $p
	}
	debugMsg "make_predicates ($unif): ($ev)($result)"
	return [list $ev $result]
}

proc get_free_variables {kind chan ev} {
	set result {}
	set free_vars [mr_free_variables $kind $chan $ev]
	# debugMsg "get_free_variables ($kind)($chan)($ev)($free_vars)"
	foreach fv $free_vars {
		lappend result [lindex $fv 0]
	}
	return $result
}

proc bindings_prepare {bindings} {
	global nb
	foreach b $bindings {
		set nb([lindex $b 0]) [lindex $b 1]
	}
}
proc bindings_update {kind chan ev} {
	global nb
	set res {}
	set free_vars [mr_free_variables $kind $chan $ev]
	# debugMsg "get_free_variables ($kind)($chan)($ev)($free_vars)"
	foreach fv $free_vars {
		set var [lindex $fv 0]
		if {[info exists nb($var)]} {
			lappend res [list [lindex $fv 1] $nb($var)]
		}
	}
	return $res
}
proc bindings_clear {} {
	global nb
	catch {unset nb}
}

proc subst_variables {kind chan ev preds} {
	set res $preds
	if {[string compare $preds ""] == 0} {
		return $res
	}
	set free_vars [mr_free_variables $kind $chan $ev]
	# debugMsg "get_free_variables ($kind)($chan)($ev)($free_vars)"
	foreach fv $free_vars {
		set cvar [lindex $fv 0]
		regsub -all {\$} $cvar {\\$} svar
		set evar [lindex $fv 1]
		# debugMsg "subst_variables try ($kind $chan $ev) ($cvar)=($svar) -> ($evar) in ($res)"
		regsub -all "($svar)(\[^0-9\])" $res "$evar\\2" res
		regsub -all "($svar)\$" $res "$evar" res
	}
	# debugMsg "subst_variables ($kind $chan $ev) ($preds) -> ($res)"
	return $res
}
