#! @EXPECT@ --
#! /Utils/bin/tclsh
#! /usr/local/bin/expect --

# NOTE: this script only needs tcl(sh); it does not use expect
# features. The only reason that we use the expect interpreter
# here is that the configure script of TorX finds the location of
# expect already for us, and (AFAIK) does not (yet?) look for tclsh.

set color(mode,init,cur) red
set color(mode,trans,cur) red2
set color(mode,init,prev) grey
set color(mode,trans,prev) grey

set cur {}

proc parse {s} {
	set res(node,init) {}
	set res(node,trans) {}
	set res(edge,init) {}
	set res(edge,trans) {}
	set words [split $s]
	set i 0
	set kind node
	set mode init
	foreach w $words {
		switch -exact -- $w {
			 init {
				set mode init
			} trans {
				set mode trans
			} edge {
				set kind edge
			} node {
				set kind node
			} default {
				lappend res($kind,$mode) $w
			}
		}
	}
	return [list $res(node,init) $res(node,trans) $res(edge,init) $res(edge,trans)]
}

# work around problem in expect 5.42,
# where stdin was made non-blocking by default
# (we could instead just use tclsh instead of expect) 
fconfigure stdin -buffering line \
                         -translation {binary binary} \
                         -blocking 1

while {[gets stdin line] >= 0} {
	set s ""

	if {[llength [lindex $cur 0]] > 0} {
		append s " -c $color(mode,init,prev)"
		append s " -n "
		append s [join [lindex $cur 0] " "]
	}
	if {[llength [lindex $cur 2]] > 0} {
		append s " -c $color(mode,init,prev)"
		append s " -e "
		append s [join [lindex $cur 2] " "]
	}
	if {[llength [lindex $cur 1]] > 0} {
		append s " -c $color(mode,trans,prev)"
		append s " -n "
		append s [join [lindex $cur 1] " "]
	}
	if {[llength [lindex $cur 3]] > 0} {
		append s " -c $color(mode,trans,prev)"
		append s " -e "
		append s [join [lindex $cur 3] " "]
	}

	set cur [parse [string trim $line]]

	if {[llength [lindex $cur 0]] > 0} {
		append s " -c $color(mode,init,cur)"
		append s " -n "
		append s [join [lindex $cur 0] " "]
	}
	if {[llength [lindex $cur 2]] > 0} {
		append s " -c $color(mode,init,cur)"
		append s " -e "
		append s [join [lindex $cur 2] " "]
	}
	if {[llength [lindex $cur 1]] > 0} {
		append s " -c $color(mode,trans,cur)"
		append s " -n "
		append s [join [lindex $cur 1] " "]
	}
	if {[llength [lindex $cur 3]] > 0} {
		append s " -c $color(mode,trans,cur)"
		append s " -e "
		append s [join [lindex $cur 3] " "]
	}

	if {[catch {puts $s} msg]} {
		catch {puts stderr "jararacy2anifsm: $msg"}
		exit
	}
}
