#! /bin/sh
# the next line restarts using wish \
exec @WISH@ "$0" -- ${1+"$@"}

# we deal here with multiple scrollbar animation runs,
# in multiple windows.
# we use 'nr' to refer to a window
# we use 'id' to refer to an animation run

wm withdraw .

set PID [pid]:

set color(background) white
set quit 0

set debug 0

set id 0
set ncanvas 0
set windows {}
set reuseQ {}

proc ldelete {l e} {
	 set ix [lsearch -exact $l $e]
	 if {$ix >= 0} {
	 	return [lreplace $l $ix $ix]
	 } else {
	 	return $l
	 }
}

proc isinteger { s } {
	# ideal implementation: return [string is integer -strict $s]
	# but that does not work in older wish versions
	if {[regexp "^\[-+\]?\[0-9\]\[0-9\]*$" $s]} {
		return 1
	} else {
		return 0
	}
}

proc isreal { s } {
	# ideal implementation: return [string is real -strict $s]
	# but that does not work in older wish versions
	if {[regexp "^\[-+\]?(\[0-9\]*\.)?\[0-9\]\[0-9\]*$" $s]} {
		return 1
	} else {
		return 0
	}
}


proc find_reusable {id ofid} {
	global reuseQ

	debug 1 "find_reusable $id : ($reuseQ) {"
	set res ""
	foreach w_nr $reuseQ {
		upvar #0 W_[set w_nr](id) w_id
		upvar #0 G_[set w_id](ieof) w_ieof
		if {[string compare $id $w_id] == 0} {
			debug 1 "find_reusable $id: skipping $w_nr: id $id == w_id $w_id"
		} elseif {$w_ieof} {
			debug 1 "find_reusable $id: found $w_nr (containing $w_id))"
			set reuseQ [ldelete $reuseQ $w_nr]
			set res $w_nr
			break
		} else {
			debug 1 "find_reusable $id: skipping $w_nr: no eof seen"
		}
	}
	debug 1 "find_reusable $id: ($reuseQ) } ($res)"
	return $res
}

proc getcanvas {id reuse ofid} {
	debug 1 "getcanvas ($id) ($reuse)($ofid) {"
	set nr [find_reusable $id $ofid]
	if {[string compare $nr ""] != 0} {
		upvar #0 W_[set nr](id) w_id
		upvar #0 S_[set w_id](nr) w_nrs
		set res_nr [reusecanvas $w_nrs $id $reuse]
		# assert($res_nr == $nr)
	} else {
		set res_nr [mkcanvas $id $reuse]
	}
	debug 1 "getcanvas ($id) ($reuse)($ofid)($res_nr) }"
	return $res_nr
}

proc reusecanvas {nrs id reuse} {

	debug 1  "reusecanvas ($nrs) ($id) ($reuse) {"
	foreach nr $nrs {
		upvar #0 W_[set nr](doReuse) doReuse
		upvar #0 W_[set nr](id) w_id
		upvar #0 W_[set nr](c) c
		upvar #0 W_[set nr](t) t
		upvar #0 W_[set nr](step) step
		upvar #0 S_[set id](nr) w_nrs
		upvar #0 G_[set id](title) title
		upvar #0 W_[set nr](doLoop) doLoop
		upvar #0 W_[set nr](delay) delay
		upvar #0 W_[set nr](delaystr) delaystr
		upvar #0 W_[set nr](curimg) curimg
		global zfact

		debug 1  "reusecanvas ($nrs) ($id) ($reuse): $nr"
		lappend w_nrs $nr

		destroy_contents $nr
#		$c delete all
		bind $t <Configure> {}

		set w_id $id
		set doReuse $reuse
		set step ""
		set doLoop 0
		set delay "0.5"
		set delaystr $delay
		set curimg {}

		wm title $t $title
		wm deiconify $t
		raise $t
	}
	update_reuse [lindex $nrs 0]
	debug 1  "reusecanvas ($nrs) ($id) ($reuse) }"

	return $nrs
}


proc mkcanvas {id reuse} {
	global color
	global orient
	global ncanvas
	global windows
	global SCROLLBARWIDTH
	upvar #0 S_[set id](nr) w_nr
	upvar #0 G_[set id](title) title
	upvar #0 W_[set ncanvas](doReuse) doReuse
	upvar #0 W_[set ncanvas](busy) busy
	upvar #0 W_[set ncanvas](destroying) destroying
	upvar #0 W_[set ncanvas](id) w_id
	upvar #0 W_[set ncanvas](c) c
	upvar #0 W_[set ncanvas](t) t
	upvar #0 W_[set ncanvas](step) step
	upvar #0 W_[set ncanvas](delay) delay
	upvar #0 W_[set ncanvas](delaystr) delaystr
	upvar #0 W_[set ncanvas](doLoop) doLoop
	upvar #0 W_[set ncanvas](curimg) curimg

	# we use busy and destroying to synchronise with the destroy
	# to avoid trying to write to a window that is just destroyed
	set busy 0
	set destroying 0

	set nr $ncanvas
	incr ncanvas

	lappend w_nr $nr
	set w_id $id
	set doReuse $reuse

	set step "0"
	set doLoop 0
	set delay "0.5"
	set delaystr $delay
	set curimg {}

	set t [toplevel .f$nr -background $color(background)]
	lappend windows $t

	# make frames first, below everything else
	frame $t.f
	frame $t.fc \
		-background  $color(background) \
		-borderwidth 0 \
		-highlightthickness 0
	frame $t.fq -borderwidth 2 -relief ridge
	frame $t.bsfn -borderwidth 2 -relief ridge
	frame $t.spfn -borderwidth 2 -relief ridge
	frame $t.bs
	frame $t.sp
	frame $t.media -borderwidth 2 -relief ridge

	set c [scrollbar $t.c]

	$c configure -command "view $nr 1"
	if {[string compare $orient ""] != 0} {
		$c configure -orient $orient
	}


	button $t.bq -text "Quit" \
		-padx 2 \
		-command "quit $nr"

	button $t.bc -text "Close" \
		-padx 2 \
		-command "mydestroy $nr"

	checkbutton $t.br -text "Reuse" \
		 -relief raised \
		 -variable W_[set nr](doReuse) \
		 -command "update_reuse $nr"
	update_reuse $nr

	label $t.bsl -text "Step:"
	entry $t.bst \
		 -width 5 \
		 -textvariable W_[set nr](step)
	bind $t.bst <Return> "showq $nr 1 \$W_[set nr](step)"
	button $t.bsp -image up \
		-command "step_inc $nr +1"
	button $t.bsm -image down \
		-command "step_inc $nr -1"

	label $t.spl -text "Delay:"
	entry $t.spt \
		 -width 5 \
		 -textvariable W_[set nr](delaystr)
	bind $t.spt <Return> "set_delay $nr \$W_[set nr](delaystr)"
	button $t.spp -image up \
		-command "delay_inc $nr +0.1"
	button $t.spm -image down \
		-command "delay_inc $nr -0.1"

	button $t.fwd -image [mk_image $nr play]
	$t.fwd configure \
		-command [list step_cont $nr +1 [$t.fwd cget -image]]
	button $t.rev -image [mk_image $nr rev ]
	$t.rev configure \
		-command [list step_cont $nr -1 [$t.rev cget -image]]
	button $t.end -image  [mk_image $nr end]
	$t.end configure \
		-command [list step_to $nr -1 [$t.end cget -image]]
	button $t.beg -image  [mk_image $nr beg]
	$t.beg configure \
		-command [list step_to $nr 0 [$t.beg cget -image]]
	button $t.stop -image  [mk_image $nr stop]
	$t.stop configure \
		-command [list step_halt $nr [$t.stop cget -image]]
	button $t.pause -image  [mk_image $nr pause]
	$t.pause configure \
		-command [list step_halt $nr [$t.pause cget -image]]

	checkbutton $t.loop -text "Loop" \
		 -relief raised \
		 -variable W_[set nr](doLoop)

	pack append $t.media \
		$t.loop {right fillx filly } \
		$t.beg {left fillx filly } \
		$t.end {right fillx filly } \
		$t.rev {left fillx filly } \
		$t.fwd {right fillx filly } \
		$t.stop {left fillx filly } \
		$t.pause {right fillx filly }

	pack append $t.fq \
		$t.bq {right } \
		$t.bc {right } \
		$t.br {right fillx filly }

#		$t.bl {right fillx filly  } \
#
	pack append $t.bs \
		$t.bsp {top  fillx filly expand } \
		$t.bsm {bottom  fillx filly expand } \

	pack append $t.sp \
		$t.spp {top  fillx filly expand } \
		$t.spm {bottom  fillx filly expand } \

	pack append $t.bsfn \
		$t.bs {right  fillx filly } \
		$t.bst {right fillx filly expand } \
		$t.bsl {left  fillx filly }

	pack append $t.spfn \
		$t.sp {right  fillx filly } \
		$t.spt {right fillx filly expand } \
		$t.spl {left  fillx filly }

	pack append $t.f \
		$t.fq {right } \
		$t.media {right  fillx filly } \
		$t.spfn {right  fillx filly expand} \
		$t.bsfn {right  fillx filly expand}

	pack append $t.fc \
		$c {top expand fill}

	pack append $t \
		$t.f {bottom fillx} \
		$t.fc {top expand fill}

	wm protocol $t WM_DELETE_WINDOW "mydestroy $nr"
	wm protocol $t WM_SAVE_YOURSELF "mydestroy $nr"
	wm title $t $title
	bind $t <q>  "quit $nr"
	bind $t <Q>  "quit $nr"


	tkwait visibility $c
	set geo [wm geometry $t]
	if {[regexp {^([0-9]+)x([0-9]+).*} $geo match w h]} {
		wm minsize $t $w $h
	}
#	bind $t <Configure>  "$c redisplay"
	return $nr

}

proc update_reuse {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 W_[set nr](doReuse) doReuse
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) w_nrs
	global reuseQ

	debug 1 "update_reuse $nr { doReuse $doReuse ieof $ieof reuseQ $reuseQ"
	foreach w_nr $w_nrs {
		upvar #0 W_[set w_nr](doReuse) w_doReuse
		set w_doReuse $doReuse
		debug 1 "update_reuse $w_nr : doReuse $doReuse ieof $ieof reuseQ $reuseQ"
		 if {$ieof && $doReuse && [lsearch -exact  $reuseQ $w_nr] < 0} {
		 	lappend reuseQ $w_nr
		 } elseif {!$doReuse} {
		 	set reuseQ [ldelete $reuseQ $w_nr]
		 }
		 if {$ieof} {
		 	.f$w_nr.br configure -state normal
		 } else {
		 	.f$w_nr.br configure -state disabled
		 }
	}
	debug 1 "update_reuse $nr } reuseQ $reuseQ"
}

proc step_inc {nr inc} {
	upvar #0 W_[set nr](id) id
	upvar #0 W_[set nr](c) c
	upvar #0 G_[set id](cur) cur
	upvar #0 G_[set id](max) max
	
	if {([string compare $cur ""] != 0) &&
	    ([isinteger $cur]) &&
	    ([expr $cur + $inc] >= 0) && ([expr $cur + $inc] <= $max)} {
		set idx $cur
	    	incr idx $inc
	    	showq $nr 1 $idx
	}
}

proc delay_inc {nr inc} {
	upvar #0 W_[set nr](delay) delay
	upvar #0 W_[set nr](delaystr) delaystr

	if {![isreal $delaystr]} {
		set d $delay
	} else {
		set d $delaystr
	}
	set s [expr $d + $inc]
	if {$s >= 0} {
		set d $s
	} else {
		set d 0
	}
	set delay $d
	set delaystr $delay
}
proc  set_delay {nr val} {
	upvar #0 W_[set nr](delay) delay
	upvar #0 W_[set nr](delaystr) delaystr

	if {![isreal $val]} {
		set d $delay
	} else {
		set d $val
	}
	set s $val
	if {$s >= 0} {
		set d $s
	} else {
		set d 0
	}
	set delay $d
	set delaystr $delay
}
proc reset_img_color {nr {img {}}} {
	upvar #0 W_[set nr](curimg) curimg

	if {$curimg != {}} {
		$curimg configure -foreground black
		set curimg {}
	}
	if {$img != {}} {
		$img  configure -foreground red
		set curimg $img
	}
}
proc step_to {nr idx img} {
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](max) max

	set_after $nr 0
	reset_img_color $nr $img
	if {$idx != 0} {
		set idx $max
	}
	showq $nr 1 $idx
	after 150 [list reset_img_color $nr]
}
proc step_cont {nr inc img} {
	reset_img_color $nr $img
	set_after $nr $inc
}
proc step_halt {nr img} {
	set_after $nr 0
	reset_img_color $nr $img
	after 150 [list reset_img_color $nr]
}

proc do_after {nr inc} {
	upvar #0 W_[set nr](id) id
	upvar #0 W_[set nr](doLoop) doLoop
	upvar #0 G_[set id](cur) cur
	upvar #0 G_[set id](max) max

	set new [expr $cur + $inc]
	if {$new >= 0 && $new <= $max} {
		step_inc $nr $inc
	} elseif {$inc > 0 && $new > $max && $doLoop} {
		set idx [expr $new - ( 1 + $max)]
		showq $nr 1 $idx
	} elseif {$inc > 0 && $new > $max && $cur < $max} {
		set idx $max
		showq $nr 1 $idx
	} elseif {$inc < 0 && $new < 0 && $doLoop} {
		set idx [expr $new + (1 + $max)]
		showq $nr 1 $idx
	} elseif {$inc < 0 && $new < 0 && $cur > 0} {
		set idx 0
		showq $nr 1 $idx
	}
	set_after $nr $inc
}

proc set_after {nr inc} {
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](cur) cur
	upvar #0 G_[set id](max) max
	upvar #0 W_[set nr](delay) delay
	upvar #0 G_[set id](after) after
	upvar #0 W_[set nr](doLoop) doLoop

	catch {after cancel $after}
	set d [expr int($delay * 1000)]
	if {$d > 0 && $doLoop} {
		set after [after $d [list do_after $nr $inc]]
	} elseif {$d > 0 && (!$doLoop && $inc > 0 && $cur < $max)} {
		set after [after $d [list do_after $nr $inc]]
	} elseif {$d > 0 && (!$doLoop && $inc < 0 && $cur > 0)} {
		set after [after $d [list do_after $nr $inc]]
	} else {
		reset_img_color $nr
	}
}

proc view {nr tell args} {
	upvar #0 W_[set nr](c) c
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](cur) cur
	upvar #0 G_[set id](max) max

	if {[llength $args] > 1 } {
		set cmd [lindex $args 0]
		if {[string compare $cmd scroll] == 0} {
			set v [expr $cur + [lindex $args 1]]
		} elseif {[string compare $cmd moveto] == 0} {
			set v [expr round( ($max + 1) * [lindex $args 1] ) ]
		}
	} else {
		set v [lindex $args 0]
	}
	debug 1 "view args=($args) v=$v"
	showq $nr $tell $v
}

proc showq {nr tell snr} {
	upvar #0 W_[set nr](c) c
	upvar #0 W_[set nr](id) id
	upvar #0 G_[set id](cur) cur
	upvar #0 G_[set id](max) max
	upvar #0 W_[set nr](step) step

	debug 1 "showq ($nr)($tell)($snr)($step)max=$max"
	if {(! $tell) && ($snr > $max)} {
		set max $snr
	}
	if {($snr >= 0) && ($snr <= $max) && ($snr != $cur)} {
#		$c  set [expr $max + 1] 1 $cur $cur
		set cur $snr
		set step $snr
		$c  set [expr $max + 1] 1 $cur $cur
		if {$tell} {
			mcastsend $id $cur
		}
	}
}


proc cmdinterp { id line chan argl } {
	upvar #0 S_[set id](nr) w_nr

	debug 1 "cmdinterp ($line) ($argl)"
	if {[isinteger  [lindex $line 0]]} {
		showq [lindex $w_nr 0] 0 [lindex $line 0]
	}
	return 1
}

proc cmdcleanup { id chan argl } {
	upvar #0 G_[set id](ifid) ifid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) w_nr

	if {[string compare $ifid $chan] == 0} {
		set ieof 1
		foreach nr $w_nr {
			update_reuse $nr
		}
	}
}



proc mcastconnect {mcast ofid } {
	set cfid {}
	set fields [split $mcast "!"]
#puts stderr "mcastconnect $fields"
	if {([llength $fields] == 3) &&
		([string compare [lindex $fields 0] tcp] == 0)} {
		set host [lindex $fields 1]
		set port [lindex $fields 2]
		if {[catch {socket $host $port} foo]} {
			warn $ofid "cannot connect to mcast $mcast: $foo"
		} else {
			msg $ofid "connected to mcast $mcast: $foo"
			set cfid $foo
		}
	} else {
		warn $ofid "cannot parse mcast \"$mcast\", expected: tcp!host.or.ip!port"
	}
	return $cfid
}

proc mcastinterp { id line chan argl } {
	upvar #0 S_[set id](nr) w_nr

	debug 1 "mcastinterp ($line) ($argl)"
	set g [lindex $argl 0]
	if {[isinteger  [lindex $line 0]]} {
		showq [lindex $w_nr 0] 0 [lindex $line 0]
	}
	return 1
}

proc mcastcleanup { id chan argl } {
	upvar #0 G_[set id](cfid) cfid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) nr

	if {[string compare $cfid $chan] == 0} {
		set cfid ""
	}
}

proc mcastsend {id nr} {
	upvar #0 G_[set id](cfid) cfid

# puts stderr "mcastsend $id $nr $cfid"
	if {[string compare $cfid ""] != 0} {
		if {[catch {output $cfid $nr} msg]} {
			debug 1 "mcastsend $nr -> $msg"
		} else {
			debug 1 "mcastsend $nr OK?"
		}
	}
}

proc debug { lvl s } {
	global debug
	if {$debug >= $lvl } {
		catch {
			puts stderr "DEBUG: mctrlsrv: $s"
			flush stderr
		}
	}
}
proc warn { fid s } {
	global PID
	catch {
		puts $fid "mctrlsrv:$PID warning: $s"
		flush $fid
	}
}
proc output { c s } {
	if {[catch {
		puts $c $s
		flush $c
	} msg]} {
		catch {puts stderr "output: $msg"}
	}
}

proc msg { fid s } {
	global PID
	catch {puts $fid "mctrlsrv:$PID $s" }
}

proc handle { chan info hproc cproc id args } {
	fileevent $chan readable {}
    if {[catch {gets $chan line} read]} {
		$cproc $id $chan $args
   } else {
	debug 2 "handle $info $hproc $cproc read: $read"
	if {$read < 0} {
		if {[eof $chan]} {
			debug 1 "Connection closed by client $info."
			 $cproc $id $chan $args
		} elseif {[fblocked $chan]} {
			warn stderr "handle $info $hproc $cproc blocked: $line"
			fileevent $chan readable \
				[concat handle $chan $info $hproc $cproc $id $args ]
		} else {
			warn stderr "handle $info error?: $line"
			$cproc $id $chan $args
		}
	} else {
		debug 3 "handle $info: $line"
		set res [$hproc $id $line $chan $args] 
		if {$res > 0} {
			fileevent $chan readable \
				[concat handle $chan $info $hproc $cproc $id $args]
		} elseif {$res < 0} {
			catch {close $chan}
			$cproc $id $chan $args
		}
	}
    }
}

proc quit {nr} {
	global B

	msg stderr "exiting"
	if {([string compare $B(interface,name) ""] != 0) &&
	    $B(interface,cleanup) &&
	    [file readable $B(interface,name)]} {
		catch {file delete -force $B(interface,name)}
	}
	if {([string compare $B(pidsfile,name) ""] != 0) &&
	    $B(pidsfile,cleanup)} {
		catch {file delete -force $B(pidsfile,name)}
	}
	exit
}
proc mydestroy {nr} {
	upvar #0 W_[set nr](busy) busy
	upvar #0 W_[set nr](destroying) destroying

	debug 1  "mydestroy $nr {"

	set destroying 1
	if {$busy} {
		debug 1  "mydestroy $nr waiting for $busy"
		trace variable busy wu "destroy_when_not_busy $nr"
	} else {
		destroy_window_and_contents $nr
	}
	debug 1  "mydestroy $nr }"
}
proc destroy_when_not_busy {nr args} {
	debug 1  "destroy_when_not_busy $nr $args {"
	destroy_window_and_contents $nr
	debug 1  "destroy_when_not_busy $nr $args }"
}
proc destroy_window_and_contents {nr} {
	global windows
	upvar #0 W_[set nr](t) t

	debug 1  "destroy_window_and_contents $nr {"
	destroy $t
	set windows [ldelete $windows $t]
	destroy_contents $nr
	cleanupnr $nr
	if {[llength $windows] <= 0} {
		quit $nr
	}
	debug 1  "destroy_window_and_contents $nr }"
}
proc destroy_contents {nr} {
	debug 1  "destroy_contents $nr {"
	cleanupid $nr
	debug 1  "destroy_contents $nr }"
}
proc cleanupid {nr} {
	upvar #0 W_[set nr](id) id
	upvar #0 S_[set id](nr) w_nr

	debug 1 "cleanup $nr {"
	set w_nr [ldelete $w_nr $nr]
	if {[llength $w_nr] == 0} {
		upvar #0 G_[set id](cfid) cfid
		upvar #0 G_[set id](ifid) ifid
		global reuseQ

		debug 1 "cleanup $nr : id $id cfid $cfid ifid $ifid"
		if {[info exists cfid] && ([string compare $cfid ""] != 0)} {
			fileevent $cfid readable {}
			close $cfid
			set cfid ""
		}
		if {[info exists ifid] && ([string compare $ifid ""] != 0)} {
			fileevent $ifid readable {}
			close $ifid
			set ifid ""
		}
	}
	debug 1 "cleanup $nr }"
}
proc cleanupnr {nr} {
	global reuseQ
	upvar #0 W_[set nr](id) id
	upvar #0 S_[set id](nr) w_nrs

	debug 1 "cleanupnr $nr: reuseQ ($reuseQ) {"
	set w_nrs [ldelete $w_nrs $nr]
	set reuseQ [ldelete $reuseQ $nr]
	catch {unset W_[set nr]}
	if {[llength $w_nrs] <= 0} {
		catch {unset S_[set id]}
	}
	
	debug 1 "cleanup $nr -> reuseQ ($reuseQ) }"
}

proc cleanupchan { id chan } {
	upvar #0 G_[set id](ifid) ifid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 S_[set id](nr) nr

	debug 1  "cleanupchan $id $chan {"
	if {[string compare $ifid $chan] == 0} {
		set ieof 1
		update_reuse $nr
	}
	debug 1  "cleanupchan $id $chan }"
}

proc setup_B {defAddr defPort fname host} {
	global B

	set B(interface,name) $fname
	set B(pidsfile,name) "$fname.pid"
	set B(statusfile,name) "$fname.stat"
	set B(interface,cleanup) 0
	set B(pidsfile,cleanup) 0
	set pid [pid]

	if {[try_connect $B(interface,name) 0]} {
		warn stderr "not replacing running server (aborting)"
		exit 0
	} else {
		# warn stderr "could not connect to running server"
	}
	if {[catch {open $B(statusfile,name) a} fid]} {
		warn stderr "cannot open (append) $B(statusfile,name)"
		exit 1
	} else {
		puts $fid "starting $host $pid"
		close $fid
	}
	if {[catch {open $B(statusfile,name) r} fid]} {
		warn stderr "cannot open (read) $B(statusfile,name)"
		exit 1
	} elseif {[gets $fid line] < 0} {
		warn stderr "cannot read $B(statusfile,name)"
		exit 1
	} else {
		close $fid
		set words [split $line]
		set stat [lindex $words 0]
		set h [lindex $words 1]
		set p [lindex $words 2]
		if {[string compare $stat starting] == 0 &&
		    [string compare $host $h] == 0 &&
		    [string compare $pid $p] != 0} {
			set found 0
			if {[catch {open "|ps -ef"} pidsfd]} {
				warn stderr "could not run: ps -ef"
				exit 1
			} else {
				while {[gets $pidsfd line] >= 0} {
					debug 25 "$pid: looking for $p in: $line"
					if {[lindex $line 1] == $p}  {
						msg stderr "found running server pid: $p"
						set found 1
						break
					}
				}
				close $pidsfd
			}
			if {$found} {
				set i 0
				set maxtries 10
				while {$i < $maxtries} {
					if {$i == ($maxtries - 1)} {
						set sock [try_connect $B(interface,name) 1]
					} else {
						set sock [try_connect $B(interface,name) 0]
					}
					if {$sock} {
						warn stderr "attempt $i: not replacing running server (aborting)"
						exit 0
					} elseif {$i == ($maxtries - 1)} {
						warn stderr "attempt $i: could not connect to running server"
					}
					incr i
				}
			} else {
				msg stderr "did not find running server pid: $p"
			}
		}
	}
	if {[try_connect $B(interface,name) 0]} {
		warn stderr "not replacing running server (aborting)"
		exit 0
	} else {
		# warn stderr "could not connect to running server"
	}
	if [catch {open $B(interface,name) w} fid] {
		warn stderr "cannot create $B(interface,name): $fid"
		return 0
	}
	set B(interface,cleanup) 1

	# msg stderr "trying to listen on tcp!$defAddr!$defPort"
	if {$defAddr == {}} {
			set c [socket -server Baccept $defPort]
		} else {
			set c [socket -server Baccept -myaddr $defAddr $defPort]
		}
	set cfg [fconfigure $c -sockname]
	set ipNr [lindex $cfg 0]
	set hostName [lindex $cfg 1]
	set portNr [lindex $cfg 2]
	msg stderr "listening on tcp!$ipNr!$portNr"
	
	puts $fid "$ipNr $hostName $portNr"
	close $fid
	
	if {![catch {open "$B(pidsfile,name)" r} pfid]} {
		while {[gets $pfid line] >= 0} {
			set p [string trim $line]
			msg stderr "cleaning up: kill $p"
			catch {exec kill $p}
			msg stderr "cleaning up: kill -9 $p"
			catch {exec kill -9 $p}
		}
		catch {close $pfid}
	}
	catch {file delete -force $B(pidsfile,name)}
	if {[catch {open "$B(pidsfile,name)" w} pfid]} {
		warn stderr "cannot open $B(pidsfile,name) for writing: $pfid"
	} else {
		set B(pidsfile,cleanup) 1
		puts $pfid $pid
		catch {close $pfid}
	}
	catch {file delete -force $B(statusfile,name)}
	return 1
}

proc Baccept {socket remoteHost remotePort} {
	global B
	
    set nSock      $socket
 	set B(busy,$socket) 0
	fconfigure $socket -buffering line ;# -blocking 0
	fileevent $socket readable [list handle $socket <B> Binterp Bcleanup ign_id]
	msg stderr "Connection accepted from client tcp!$remoteHost!$remotePort"
}  


proc Binterp { ign_id line chan argl } {
	return [continue_reading_after_process_B_line $line $chan]
}

proc Bcleanup { ign_id chan argl } {
	global B

	if {![catch {fconfigure $chan -peername} cfg]} {
		set remoteHost [lindex $cfg 0]
		set remotePort [lindex $cfg 2]
		msg stderr "Connection closed by client tcp!$remoteHost!$remotePort"
	}
	catch {close $chan}
	unset B(busy,$chan)
}

proc continue_reading_after_process_B_line { line fid } {
	global B
	set name [string trim $line]
	set result 1
	if {([string compare $name -exit] == 0) ||
	    (([string compare [lindex $name 0] -args] == 0) &&
	     ([string compare [lindex $name 1] -exit] == 0))} {
		# exit will do this for us: cleanupchan
		quit ignored
	} elseif {[string compare [lindex $name 0] -eof] == 0} {
		debug 1 "B eof"
		trace variable B(busy,$fid) w "close_B_when_ready $fid"
		return 0
	} elseif {[string compare [lindex $name 0] -debug] == 0} {
		debug 1 "B debug: [lrange $name 1 end]"
	} elseif {[string compare [lindex $name 0] -args] == 0} {
		debug 1 "B debug: [lrange $name 1 end]"
		set result [continue_reading_after_process_args [lrange $name 1 end] $fid $fid]
	} else {
		debug 1 "B debug: unknown line: $line"
	}
	return $result
}

proc close_B_when_ready {fid varName index op} {
	global B
	debug 1 "close_B_when_ready ($fid)($varName)($index)($op)"
	switch $op {
		w {
			if {[string compare "busy,$fid" $index] == 0} {
				if {$B(busy,$fid) == 0} {
					debug 1 "close_B_when_ready: closing fid (B(busy,$fid) == 0)"
					if {![catch {fconfigure $fid -peername} cfg]} {
						set remoteHost [lindex $cfg 0]
						set remotePort [lindex $cfg 2]
						msg stderr "Closing connection to client tcp!$remoteHost!$remotePort"
					}
					close $fid
					unset B(busy,$fid)
				} else {
					debug 1 "close_B_when_ready: not yet closing fid (B(busy,$fid) == $B(busy,$fid))"
				}
			} else {
					debug 1 "close_B_when_ready: unrelated update (index=$index)"
				
			}
		}
	}
}

proc continue_reading_after_process_args {argv ifid ofid {name stdin} } {
	global env
	global id
	global orient
	
	set reuse 0
	set files {}
	set result 1
	set mcast ""
	set title "MControl"
	set orient "horizontal"
	
	if {[info exists env(TORXMCASTID)]} {
		set mcast $env(TORXMCASTID)
	}

	set args $argv
	
	while {[llength $args] > 0} {
		set a [lindex $args 0]
		set args [lrange $args 1 end]
		switch -- $a {
		    -r {
				set reuse 1
		  } -m {
				if {[llength $args] <= 0} {
		  			warn $ofid "missing argument for flag: $a"
				} else {
					set mcast [lindex $args 0]
					set args [lrange $args 1 end]
				}
		  } -t {
				if {[llength $args] <= 0} {
		  			warn $ofid "missing argument for flag: $a"
				} else {
					set title "MControl"
					append title " - "
					append title [lindex $args 0]
					set args [lrange $args 1 end]
				}
		  } -* {
		  		warn $ofid "unknown flag: $a"
				set result -1
		  } default {
		  		lappend files $a
		  }
		}
	}

	if {($result == -1) || ($ifid == {}) || ([llength $files] != 0)} {
		warn $ofid "usage: mctrl \[-t title\] -m mcastid"
		set result -1
	} else {
		set cfid [mcastconnect $mcast $ofid]
		if {([string compare $mcast ""] == 0) || ($cfid != {})} {
			incr id
			start $id $reuse $ifid $ofid $cfid $name $title
			set result 0
		} else {
			set result -1
		}
	}
	return $result
}


proc start {id rse aifid aofid acfid name atitle} {
	global idinfo
	upvar #0 G_[set id](ifid) ifid
	upvar #0 G_[set id](ieof) ieof
	upvar #0 G_[set id](iname) iname
	upvar #0 G_[set id](ofid) ofid
	upvar #0 G_[set id](cfid) cfid
	upvar #0 G_[set id](busy) busy
	upvar #0 G_[set id](destroying) destroying
	upvar #0 G_[set id](reuse) reuse
	upvar #0 G_[set id](title) title
	upvar #0 G_[set id](cur) cur
	upvar #0 G_[set id](max) max
	global env

	set reuse $rse
	set ifid $aifid
	set ieof 0
	set iname $name
	set ofid $aofid
	set cfid $acfid
	set title $atitle

	# we use busy and destroying to synchronise with the destroy
	# to avoid trying to write to a window that is just destroyed
	set busy 0
	set destroying 0

	set idinfo($id,started) 1
	set idinfo($id,closed) 0

	if {[info exists B(busy,$ofid)]} {
		incr B(busy,$ofid)
	}
	
	if {[string compare $ifid stdin ] != 0} {
		fconfigure $ifid -buffering line ;# -blocking 0
	}

	set nr [getcanvas $id $reuse $ofid]
	set max -1
	set cur -1
	showq $nr 0 0

	fileevent $ifid readable [list handle $ifid $name cmdinterp cmdcleanup $id]
	debug 1 "start: started cmdinterp $id $ifid $name"
	if {[string compare $cfid ""] != 0} {
		fileevent $cfid readable [list handle $cfid "<mcast>" mcastinterp mcastcleanup $id ]
	}

}

# =================================

proc find_user {host} {
	global env

	if {[info exists env(USER)]} {
		set user $env(USER)
	} elseif {[info exists env(LOGNAME)]} {
		set user $env(LOGNAME)
	} else {
		set user unknown
	}
	return $user
}

proc find_display {host} {
	global env

	if {[info exists env(DISPLAY)]} {
		set display $env(DISPLAY)
	} else {
		set display ":0"
	}
	regsub {^:} $display "$host:" display
	regsub {^unix:} $display "$host:" display
	regsub {\.[0-9]*$} $display "" display
	return $display
}

proc mk_filename {dir pfx user display} {
	set fname [file join $dir "$pfx-$user-$display"]
	regsub -all : $fname _ fname
	return $fname
}

proc try_connect { fname do_warn } {
	if {![catch {open $fname r} fid]} {
		if {[gets $fid line] > 0} {
			set words [split $line]
			set ipNr [lindex $words 0]
			set hostName [lindex $words 1]
			set portNr [lindex $words 2]
			if {![catch {socket $ipNr $portNr} socket]} {
				return 1
			} elseif {$do_warn} {
				warn stderr "could not connect tcp!$ipNr!$portNr: $socket"
			}
		}
		close $fid
	} elseif {$do_warn} {
		warn stderr "$fid"
	}
	return 0
}

proc mk_buttons {} {

	image create bitmap up -data {
#define up_width 5
#define up_height 3
static unsigned char up_bits[] = {
   0x04, 0x0e, 0x1f};
	}
	image create bitmap down -data {
#define down_width 5
#define down_height 3
static unsigned char down_bits[] = {
   0x1f, 0x0e, 0x04};
	}
}

proc mk_image {nr type} {
	switch -exact -- $type {
		play {
			return [image create bitmap play_$nr -data {
#define play_width 13
#define play_height 13
static unsigned char play_bits[] = {
   0x00, 0x00, 0x18, 0x00, 0x38, 0x00, 0x78, 0x00, 0xf8, 0x00, 0xf8, 0x01,
   0xf8, 0x03, 0xf8, 0x01, 0xf8, 0x00, 0x78, 0x00, 0x38, 0x00, 0x18, 0x00,
   0x00, 0x00};
			} ]
		}
		rev {
			return [image create bitmap rev_$nr -data {
#define rev_width 13
#define rev_height 13
static unsigned char rev_bits[] = {
   0x00, 0x00, 0x00, 0x03, 0x80, 0x03, 0xc0, 0x03, 0xe0, 0x03, 0xf0, 0x03,
   0xf8, 0x03, 0xf0, 0x03, 0xe0, 0x03, 0xc0, 0x03, 0x80, 0x03, 0x00, 0x03,
   0x00, 0x00};
			} ]
		}
		end {
			return [image create bitmap end_$nr -data {
#define end_width 13
#define end_height 13
static unsigned char end_bits[] = {
   0x00, 0x00, 0x0c, 0x06, 0x1c, 0x06, 0x3c, 0x06, 0x7c, 0x06, 0xfc, 0x06,
   0xfc, 0x07, 0xfc, 0x06, 0x7c, 0x06, 0x3c, 0x06, 0x1c, 0x06, 0x0c, 0x06,
   0x00, 0x00};
			} ]
		}
		beg {
			return [image create bitmap beg_$nr -data {
#define beg_width 13
#define beg_height 13
static unsigned char beg_bits[] = {
   0x00, 0x00, 0x0c, 0x06, 0x0c, 0x07, 0x8c, 0x07, 0xcc, 0x07, 0xec, 0x07,
   0xfc, 0x07, 0xec, 0x07, 0xcc, 0x07, 0x8c, 0x07, 0x0c, 0x07, 0x0c, 0x06,
   0x00, 0x00};
			} ]
		}
		stop {
			return [image create bitmap stop_$nr -data {
#define stop_width 13
#define stop_height 13
static unsigned char stop_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0xfc, 0x07, 0xfc, 0x07, 0xfc, 0x07, 0xfc, 0x07,
   0xfc, 0x07, 0xfc, 0x07, 0xfc, 0x07, 0xfc, 0x07, 0xfc, 0x07, 0x00, 0x00,
   0x00, 0x00};
			} ]
		}
		pause {
			return [image create bitmap pause_$nr -data {
#define pause_width 13
#define pause_height 13
static unsigned char pause_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0xb8, 0x03, 0xb8, 0x03, 0xb8, 0x03, 0xb8, 0x03,
   0xb8, 0x03, 0xb8, 0x03, 0xb8, 0x03, 0xb8, 0x03, 0xb8, 0x03, 0x00, 0x00,
   0x00, 0x00};
			} ]
		}
	}
}

proc main {} {
	global argv
	
	mk_buttons

	if {[llength $argv] > 0} {
		set s [join $argv " "]
		warn stderr "mctlrsrv: ignoring arguments: $s"
		warn stderr "usage: mctlrsrv"
	}
	set host [exec uname -n]
	set display [find_display $host]
	set user [find_user $host]
	set fname [mk_filename /tmp mctrl $user $display]

	if {![setup_B {} 0 $fname $host]} {
		exit 1
	}
	
#	if {[continue_reading_after_process_args $argv stdin stderr] < 0} {
#		# hmmm... is this what we want,
#		# or should server remain running?
#		quit ignored
#	}
}

main

vwait quit
exit
